from . import docs_general


def docstring(header, reduction=None):
    def arguments():
        return r"""
Arguments
---------
data: Tuple[torch.Tensor, torch.Tensor]
    Tuple containing `outputs` from neural network and regression `targets`.
    `outputs` should be of shape :math:`(N, *)`, where :math:`N` is the number of samples.
    Should contain `floating` point values.
    `targets` should be in the same shape `outputs` and be of `float` data type as well.
        """

    def wrapper(klass):
        docstring = r"""{}

        """.format(
            header
        )
        if reduction is not None:
            docstring += r"""
Parameters
----------
            """
            docstring += docs_general.reduction_parameter(reduction)

        klass.forward.__doc__ = arguments() + docs_general.returned(reduction)
        klass.__doc__ = docstring
        return klass

    return wrapper



"""Quantization related operations."""

import torch

from ._base import Operation


class Dequantize(Operation):
    """Given a quantized Tensor, dequantize it and return `float` Tensor.

    Arguments
    ---------
    data: torch.Tensor
        Quantized `torch.Tensor` to dequantize.



    """

    def forward(self, data):
        """
        Returns
        ---------
        data: torch.Tensor
            Dequantized to `float32` `torch.Tensor`
        """

        return data.dequantize()



"""Utilities used with most binary metrics."""

import typing

import torch


def check(output: torch.Tensor, target: torch.Tensor) -> None:
    """
    Check `output` and `target` shape is equal.

    ValueError is raised otherwise

    Parameters
    ----------
    output : torch.Tensor
        Usually neural network output
    target : torch.Tensor
        Usually desired target

    """

    if output.shape != target.shape:
        raise ValueError(
            "Output and target has to be of the same shape! Got {} for output and {} for target".format(
                output.shape, target.shape
            )
        )


def threshold(
    output: torch.Tensor, target: torch.Tensor, threshold: float
) -> typing.Tuple[torch.Tensor, torch.Tensor]:
    """
    Threshold output logits/probabilities at `thresholds`.

    For logits, `threshold` should be set to `0`, while for `probabilities`
    it should be `0.5`

    Parameters
    ----------
    output : torch.Tensor
        Usually neural network output
    target : torch.Tensor
        Usually desired target
    threshold : float
        Value at which `output`will be thresholded

    Returns
    -------
    typing.Tuple[torch.Tensor, torch.Tensor]:
        Thresholded output and `target` being a boolean for faster computations.
    """
    return output > threshold, target.bool()



import itertools
import operator
import os
import pathlib
import shutil
import tempfile

import pytest
import torch
import torchfunc
import torchtraining as tt


@pytest.mark.parametrize(
    "iterations,comparator", itertools.product((1, 5), (operator.ge, operator.le))
)
def test_save(iterations, comparator):
    if comparator == operator.ge:
        files = iterations
    else:
        files = 1

    directory = pathlib.Path(tempfile.mkdtemp())
    module = torch.nn.Linear(20, 1)
    save = tt.callbacks.Save(module, directory / "module.py", comparator=comparator)

    for value in range(iterations):
        save.path = directory / "{}.pt".format(value)
        save(value)

    file_count = sum(len(files) for _, _, files in os.walk(directory))
    assert file_count == files
    shutil.rmtree(directory)


def test_time_stopping():
    with pytest.raises(tt.exceptions.TimeStopping) as e_info:
        stopper = tt.callbacks.TimeStopping(1)
        while True:
            stopper(1)


@pytest.mark.parametrize("iterations,nan", itertools.product((1, 5), (True, False)))
def test_terminate_nan(iterations, nan):
    if nan:
        inputs = torch.tensor([1, float("nan"), 2])
        expected = 0
    else:
        inputs = torch.ones(10)
        expected = iterations

    value = 0
    try:
        terminator = tt.callbacks.TerminateOnNan()
        for _ in range(iterations):
            terminator(inputs)
            value += 1
    except:
        pass

    assert expected == value


@pytest.mark.parametrize(
    "n,iterations,freeze", itertools.product((0, 10), (5, 10), (True, False)),
)
def test_unfreeze(n, iterations, freeze):
    expected = n < iterations
    module = torch.nn.Linear(20, 10)
    unfreezer = tt.callbacks.Unfreeze(module, n)
    if freeze:
        torchfunc.module.freeze(module)

    for i in range(iterations):
        unfreezer(i)

    for param in module.parameters():
        assert param.requires_grad == expected or not freeze



import torchtraining as tt


def accuracy_pipeline(iteration, writer, name: str):
    return (
        iteration
        ** tt.Select(predictions=1, labels=2)
        ** tt.metrics.classification.multiclass.Accuracy()
        ** tt.accumulators.Mean()
        ** tt.Split(
            tt.callbacks.Log(f"{name} Accuracy"),
            tt.callbacks.tensorboard.Scalar(writer, f"{name}/Accuracy"),
        )
    )


def loss_pipeline(iteration, writer, name: str):
    return (
        iteration
        ** tt.Select(loss=0)
        ** tt.accumulators.Mean()
        ** tt.Split(
            tt.callbacks.Log(f"{name} Loss"),
            tt.callbacks.tensorboard.Scalar(writer, f"{name}/Loss", log="INFO"),
        )
    )


def train(writer, step, module, data, name: str):
    iteration = tt.iterations.Train(step, module, data, log="INFO")
    return accuracy_pipeline(loss_pipeline(iteration, writer, name), writer, name)


def eval(writer, step, module, data, name: str):
    iteration = tt.iterations.Eval(step, module, data, log="INFO")
    return accuracy_pipeline(loss_pipeline(iteration, writer, name), writer, name)



import abc
import collections
import typing

import torch

from ... import _base, functional
from . import utils

###############################################################################
#
#                           COMMON BASE CLASSES
#
###############################################################################


class _Threshold(_base.Operation):
    def __init__(self, threshold: float = 0.0):
        super().__init__()

        self.threshold = threshold

    @abc.abstractmethod
    def forward(self, data):
        pass


class _ThresholdReductionMean(_base.Operation):
    def __init__(self, threshold: float = 0.0, reduction=torch.mean):
        super().__init__()

        self.threshold = threshold
        self.reduction = reduction

    @abc.abstractmethod
    def forward(self, data):
        pass


class _ThresholdReductionSum(_base.Operation):
    def __init__(self, threshold: float = 0.0, reduction=torch.sum):
        super().__init__()

        self.threshold = threshold
        self.reduction = reduction

    @abc.abstractmethod
    def forward(self, data):
        pass


###############################################################################
#
#                          CONCRETE IMPLEMENTATIONS
#
###############################################################################


###############################################################################
#
#                           MEAN DEFAULT REDUCTION
#
###############################################################################


@utils.binary.docs(
    header="""Calculate accuracy score between `output` and `target`.""",
    reduction="mean",
)
class Accuracy(_ThresholdReductionMean):
    def forward(self, data):
        return functional.metrics.classification.binary.accuracy(
            *data, self.threshold, self.reduction,
        )


@utils.binary.docs(
    header="""Calculate jaccard score between `output` and `target`.""",
    reduction="mean",
)
class Jaccard(_ThresholdReductionMean):
    def forward(self, data):
        return functional.metrics.classification.binary.jaccard(
            *data, self.threshold, self.reduction,
        )


###############################################################################
#
#                           SUM DEFAULT REDUCTION
#
###############################################################################


@utils.binary.docs(
    header="""Number of true positives between `output` and `target`.""",
    reduction="sum",
)
class TruePositive(_ThresholdReductionSum):
    def forward(self, data):
        return functional.metrics.classification.binary.true_positive(
            *data, self.threshold, self.reduction
        )


@utils.binary.docs(
    header="""Number of false positives between `output` and `target`.""",
    reduction="sum",
)
class FalsePositive(_ThresholdReductionSum):
    def forward(self, data):
        return functional.metrics.classification.binary.false_positive(
            *data, self.threshold, self.reduction
        )


@utils.binary.docs(
    header="""Number of true negatives between `output` and `target`.""",
    reduction="sum",
)
class TrueNegative(_ThresholdReductionSum):
    def forward(self, data):
        return functional.metrics.classification.binary.true_negative(
            *data, self.threshold, self.reduction
        )


@utils.binary.docs(
    header="""Number of false negatives between `output` and `target`.""",
    reduction="sum",
)
class FalseNegative(_ThresholdReductionSum):
    def forward(self, data):
        return functional.metrics.classification.binary.false_negative(
            *data, self.threshold, self.reduction
        )


@utils.binary.docs(
    header="""Confusion matrix between `output` and `target`.""", reduction="sum",
)
class ConfusionMatrix(_ThresholdReductionSum):
    def forward(self, data):
        return functional.metrics.classification.binary.confusion_matrix(
            *data, self.threshold, self.reduction
        )


###############################################################################
#
#                               NO REDUCTION
#
###############################################################################


@utils.binary.docs(header="""Recall between `output` and `target`.""",)
class Recall(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.recall(*data, self.threshold)


@utils.binary.docs(header="""Specificity between `output` and `target`.""",)
class Specificity(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.specificity(
            *data, self.threshold
        )


@utils.binary.docs(header="""Precision between `output` and `target`.""",)
class Precision(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.precision(*data, self.threshold)


@utils.binary.docs(
    header="""Negative predictive value between `output` and `target`.""",
)
class NegativePredictiveValue(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.negative_predictive_value(
            *data, self.threshold
        )


@utils.binary.docs(header="""False negative rate between `output` and `target`.""",)
class FalseNegativeRate(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.false_negative_rate(
            *data, self.threshold
        )


@utils.binary.docs(header="""False positive rate between `output` and `target`.""",)
class FalsePositiveRate(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.false_positive_rate(
            *data, self.threshold
        )


@utils.binary.docs(header="""False discovery rate between `output` and `target`.""",)
class FalseDiscoveryRate(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.false_discovery_rate(
            *data, self.threshold
        )


@utils.binary.docs(header="""False omission rate between `output` and `target`.""",)
class FalseOmissionRate(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.false_omission_rate(
            *data, self.threshold
        )


@utils.binary.docs(header="""Critical success index between `output` and `target`.""",)
class CriticalSuccessIndex(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.critical_success_index(
            *data, self.threshold
        )


@utils.binary.docs(header="""Critical success index between `output` and `target`.""",)
class BalancedAccuracy(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.balanced_accuracy(
            *data, self.threshold
        )


@utils.binary.docs(header="""F1 score between `output` and `target`.""",)
class F1(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.f1(*data, self.threshold)


@utils.binary.docs(
    header="""Matthews correlation coefficient between `output` and `target`.""",
)
class MatthewsCorrelationCoefficient(_Threshold):
    def forward(self, data):
        return functional.metrics.classification.binary.matthews_correlation_coefficient(
            *data, self.threshold
        )


###############################################################################
#
#                               OTHER METRICS
#
###############################################################################


class FBeta(_base.Operation):
    r"""Get f-beta score between `outputs` and `targets`.


    Works for both logits and probabilities of `output`.

    If `output` is tensor after `sigmoid` activation
    user should change `threshold` to `0.5` for
    correct results (default `0.0` corresponds to unnormalized probability a.k.a logits).

    Parameters
    ----------
    beta: float
        Beta coefficient of `f-beta` score.
    threshold : float, optional
        Threshold above which prediction is considered to be positive.
        Default: `0.0`

    Arguments
    ---------
    data: Tuple[torch.Tensor, torch.Tensor]
        Tuple containing `outputs` from neural network and `targets` (ground truths).
        `outputs` should be of shape :math:`(N, *)` and contain `logits` or `probabilities`.
        `targets` should be of shape :math:`(N, *)` as well and contain `boolean` values
        (or integers from set :math:`{0, 1}`).

    Returns
    -------
    torch.Tensor
        Scalar `tensor`

    """

    def __init__(self, beta: float, threshold: float = 0.0):
        super().__init__()
        self.beta = beta
        self.threshold = threshold

    def forward(self, data):
        return functional.metrics.classification.binary.f_beta(
            *data, self.beta, self.threshold
        )



import itertools

import numpy as np
import torch
from sklearn.metrics import (accuracy_score, balanced_accuracy_score, f1_score,
                             fbeta_score, jaccard_score, matthews_corrcoef,
                             precision_score, recall_score)

import pytest
import torchtraining as tt
import torchtraining.metrics.classification.binary as M

# @pytest.mark.parametrize(
#     "metrics,data",
#     list(
#         itertools.product(
#             (
#                 (M.Accuracy(), accuracy_score),
#                 (M.Jaccard(), jaccard_score),
#                 (M.Recall(), recall_score),
#                 (M.Precision(), precision_score),
#                 (M.BalancedAccuracy(), balanced_accuracy_score),
#                 (M.F1(), f1_score),
#                 (
#                     M.FBeta(beta=4),
#                     lambda y_true, y_pred: fbeta_score(y_true, y_pred, beta=4),
#                 ),
#                 (M.MatthewsCorrelationCoefficient(), matthews_corrcoef),
#             ),
#             (torch.randn(10, 5), torch.randint(high=2, size=(10,))),
#         ),
#     ),
# )
# def test_binary(metrics, data):
#     package, scikit = metrics
#     outputs, targets = data
#     our = package(data).numpy()
#     their = scikit(targets.numpy(), torch.argmax(outputs, dim=-1).numpy())
#     assert np.isclose(our, their)


@pytest.mark.parametrize(
    "metric,data",
    list(
        itertools.product(
            (
                M.Accuracy(),
                M.Jaccard(),
                M.TruePositive(),
                M.FalsePositive(),
                M.FalsePositive(),
                M.TrueNegative(),
                M.FalseNegative(),
                # Special handling
                M.ConfusionMatrix(),
                M.Recall(),
                M.Specificity(),
                M.Precision(),
                M.NegativePredictiveValue(),
                M.FalseNegativeRate(),
                M.FalsePositiveRate(),
                M.FalseDiscoveryRate(),
                M.FalseOmissionRate(),
                M.CriticalSuccessIndex(),
                M.BalancedAccuracy(),
                M.F1(),
                M.FBeta(beta=2.4),
                M.MatthewsCorrelationCoefficient(),
            ),
            (
                (torch.randn(4, 10), torch.randint(high=2, size=(4, 10))),
                (torch.randn(4, 2, 5), torch.randint(high=2, size=(4, 2, 5))),
            ),
        )
    ),
)
def test_smoke(metric, data):
    metric(data)



"""Accelerators enabling distributed (multi-GPU/multi-node) training.

Accelerators should be instantiated only once and used on top-most
module (in the following order):

    * epoch (if exists)
    * iteration (if exists)
    * step

Those are the only objects which can be "piped" into producers, for example::

    tt.accelerators.Horovod(...) ** tt.iterations.Iteration(...)

And should be used in this way (although it's not always necessary).
See `horovod` module for an example.

"""

import importlib

import torch

from .._base import Accelerator
from ..utils import general as utils

if utils.modules_exist("horovod", "horovod.torch"):

    import horovod.torch as hvd

    class Horovod(Accelerator):
        """Accelerate training using Uber's Horovod framework.

        See `torchtraining.accelerators.horovod` package for more information.

        .. note::

            **IMPORTANT**: This object needs `horovod` Python package to be visible.
            You can install it with `pip install -U torchtraining[horovod]`.
            Also you should export `CUDA_HOME` variable like this:
            `CUDA_HOME=/opt/cuda pip install -U torchtraining[horovod]` (your path may vary)


        Parameters
        ----------
        module: torch.nn.Module
            Module to be broadcasted to all processes.
        rank: int, optional
            Root process rank. Default: `0`
        per_worker_threads: int, optional
            Number of threads which can be utilized by each process.
            Default: `pytorch`'s default
        comm: List, optional
            List specifying ranks for the communicator, relative to the `MPI_COMM_WORLD`
            communicator OR the MPI communicator to use. Given communicator will be duplicated.
            If `None`, Horovod will use MPI_COMM_WORLD Communicator.
            Default: `None`

        """

        def __init__(
            self, model, rank: int = 0, per_worker_threads: int = None, comm=None,
        ):
            hvd.init(comm)

            if torch.cuda.is_available():
                torch.cuda.set_device(hvd.local_rank())
            if per_worker_threads is not None:
                if per_worker_threads < 1:
                    raise ValueError("Each worker needs at least one thread to run.")
                torch.set_num_threads(per_worker_threads)

            hvd.broadcast_parameters(model.state_dict(), root_rank=rank)

    from . import horovod



import typing

import torch
from torch.nn.modules.loss import _Loss

from . import functional

# DICE


# Dice loss


class BinaryFocal(_Loss):
    """Binary focal loss working with raw output from network (logits).

    See original research paper: `Focal Loss for Dense Object Detection <https://arxiv.org/abs/1708.02002>`__

    Underplays loss of easy examples while leaving loss of harder examples
    for neural network mostly intact (dampened way less).

    The higher the gamma parameter, the greater the "focusing" effect.

    Arguments
    ---------
    gamma: float
        Scale of focal loss effect. To obtain binary crossentropy set it to 0.0.
        `0.5 - 2.5` range was used in original research paper and seemed robust.
    weight: Tensor, optional
        Manual rescaling weight, if provided it's repeated to match input
        tensor shape
    pos_weight: Tensor, optional
        Weight of positive examples. Must be a vector with
        length equal to the number of classes.
        In general `pos_weight` should be decreased slightly as `gamma` is increased
        (for `gamma=2`, `pos_weight=0.25` was found to work best in original paper).
    reduction: typing.Callable(torch.Tensor) -> torch.Tensor, optional
        Specifies the reduction to apply to the output.
        If user wants no reduction he should use: `lambda loss: loss`.
        If user wants a summation he should use: `torch.sum`.
        By default, `lambda loss: loss.sum() / loss.shape[0]` is used (mean across examples).


    """

    def __init__(
        self,
        gamma: float,
        weight=None,
        pos_weight=None,
        reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
    ):
        super().__init__()

        self.gamma = gamma
        self.weight = weight
        self.pos_weight = pos_weight
        self.reduction = reduction

    def forward(self, outputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """
        Arguments
        ---------
        outputs: torch.Tensor
            :math:`(N, *)` where :math:`*` means, any number of additional dimensions.
            Usually of shape :math:`(N, H, W)`, where :math:`H` is image height and :math:`W`
            is it's width.
        targets: torch.Tensor
            :math:`(N, *)`, same shape as the input.

        Returns
        -------
        torch.Tensor
            If :attr:`reduction` is not specified then `mean` across sample is taken.
            Otherwise whatever shape `reduction` returns.

        """
        return functional.loss.binary_focal_loss(
            outputs, targets, self.gamma, self.weight, self.pos_weight, self.reduction,
        )


class MulticlassFocal(_Loss):
    r"""Multiclass focal loss working with raw output from network (logits).

    See original research paper: `Focal Loss for Dense Object Detection <https://arxiv.org/abs/1708.02002>`__

    Underplays loss of easy examples while leaving loss of harder examples
    for neural network mostly intact (dampened way less).

    The higher the gamma parameter, the greater the "focusing" effect.

    Arguments
    ---------
    gamma: float
        Scale of focal loss effect. To obtain binary crossentropy set it to 0.0.
        `0.5 - 2.5` range was used in original research paper and seemed robust.
    weight: Tensor, optional
        Manual rescaling weight, if provided it's repeated to match input
        tensor shape.
    ignore_index int, optional
        Specifies a target value that is ignored and does not contribute to the input gradient.
        When :attr:`size_average` is ``True``, the loss is averaged over non-ignored targets.
    reduction: typing.Callable(torch.Tensor) -> torch.Tensor, optional
        Specifies the reduction to apply to the output.
        If user wants no reduction he should use: `lambda loss: loss`.
        If user wants a summation he should use: `torch.sum`.
        By default, `lambda loss: loss.sum() / loss.shape[0]` is used (mean across examples).

    """

    def __init__(
        self,
        gamma: float,
        weight=None,
        ignore_index: int = -100,
        reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
    ):
        super().__init__()

        self.gamma = gamma
        self.weight = weight
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, outputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """
        Arguments
        ---------
        outputs: torch.Tensor
            :math:`(N, C)` where `C = number of classes`, or
            :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1`
            in the case of `K`-dimensional loss.
            Usually of shape :math:`(N, C, H, W)`, where :math:`H` is image height and :math:`W`
            is it's width.
        targets: torch.Tensor
            :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or
            :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
            K-dimensional loss.
            Usually of shape :math:`(N, H, W)`, where :math:`H` is image height and :math:`W`
            is it's width and elements are of specified `C` classes.

        Returns
        -------
        torch.Tensor
            If :attr:`reduction` is not specified then `mean` across sample is taken.
            Otherwise whatever shape `reduction` returns.

        """
        return functional.loss.multiclass_focal_loss(
            outputs, targets, self.gamma, self.weight, self.ignore_index, self.reduction
        )


# Inspired by: https://stackoverflow.com/questions/55681502/label-smoothing-in-pytorch
# and adjusted to fit torchtraining API
class SmoothCrossEntropy(_Loss):
    r"""Run cross entropy with non-integer labels smoothed by `alpha`.

    See `When Does Label Smoothing Help? <https://arxiv.org/abs/1906.02629>`__ for more details

    `targets` will be transformed to one-hot encoding and modified according
    to formula:

    .. math::
        y = y(1 - \alpha) + \frac{\alpha}{C}

    where :math:`C` is total number of classes.

    Arguments
    ---------
    alpha: float
        Smoothing parameter in the range `[0, 1)`.
    weight: Tensor, optional
        Manual rescaling weight, if provided it's repeated to match input
        tensor shape. Default: `None` (no weighting)
    ignore_index int, optional
        Specifies a target value that is ignored and does not contribute to the input gradient.
        When :attr:`size_average` is ``True``, the loss is averaged over non-ignored targets.
        Default: `-100`
    reduction: typing.Callable(torch.Tensor) -> torch.Tensor, optional
        Specifies the reduction to apply to the output.
        If user wants no reduction he should use: `lambda loss: loss`.
        If user wants a summation he should use: `torch.sum`.
        By default, `lambda loss: loss.sum() / loss.shape[0]` is used (mean across examples).

    """

    def __init__(
        self,
        alpha: float,
        weight=None,
        ignore_index: int = -100,
        reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
    ):
        if not 0 <= alpha < 1:
            raise ValueError("smoothing alpha should be in [0, 1) range.")

        super().__init__()

        self.alpha = alpha
        self.weight = weight
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, outputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """
        Arguments
        ---------
        outputs: torch.Tensor
            :math:`(N, C)` where `C = number of classes`, or
            :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1`
            in the case of `K`-dimensional loss.
        targets: torch.Tensor
            :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or
            :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
            K-dimensional loss.
        Returns
        -------
        torch.Tensor
            If :attr:`reduction` is not specified then `mean` across sample is taken.
            Otherwise whatever shape `reduction` returns.

        """
        return functional.loss.smooth_cross_entropy(
            outputs, targets, self.alpha, self.weight, self.ignore_index, self.reduction
        )


class SmoothBinaryCrossEntropy(_Loss):
    r"""Run binary cross entropy with booleans smoothed by `alpha`.

    See `When Does Label Smoothing Help? <https://arxiv.org/abs/1906.02629>`__ for more details

    `targets` will be transformed to one-hot encoding and modified according
    to formula:

    .. math::
        y = y(1 - \alpha) + \frac{\alpha}{2}

    where :math:`2` is total number of classes in binary case.

    Arguments
    ---------
    alpha: float
        Smoothing parameter in the range `[0, 1)`.
    weight: Tensor, optional
        Manual rescaling weight, if provided it's repeated to match input
        tensor shape. Default: `None` (no weighting)
    pos_weight: Tensor, optional
        Weight of positive examples. Must be a vector with
        length equal to the number of classes.
        In general `pos_weight` should be decreased slightly as `gamma` is increased
        (for `gamma=2`, `pos_weight=0.25` was found to work best in original paper).
    reduction: typing.Callable(torch.Tensor) -> torch.Tensor, optional
        Specifies the reduction to apply to the output.
        If user wants no reduction he should use: `lambda loss: loss`.
        If user wants a summation he should use: `torch.sum`.
        By default, `lambda loss: loss.sum() / loss.shape[0]` is used (mean across examples).


    """

    def __init__(
        self,
        alpha: float,
        weight=None,
        pos_weight: int = None,
        reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
    ):
        if not 0 <= alpha < 1:
            raise ValueError("smoothing alpha should be in [0, 1) range.")

        super().__init__()

        self.alpha = alpha
        self.weight = weight
        self.pos_weight = pos_weight
        self.reduction = reduction

    def forward(self, outputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """
        Arguments
        ---------
        outputs: torch.Tensor
            :math:`(N, *)` where :math:`*` means, any number of additional dimensions
        targets: torch.Tensor
            :math:`(N, *)`, same shape as the input

        Returns
        -------
        torch.Tensor
            If :attr:`reduction` is not specified then `mean` across sample is taken.
            Otherwise whatever shape `reduction` returns.
        """
        return functional.loss.smooth_binary_cross_entropy(
            outputs, targets, self.alpha, self.weight, self.pos_weight, self.reduction
        )


class QuadrupletLoss(_Loss):
    r"""Quadruplet loss pushing away samples belonging to different classes.

    See original research paper `Beyond triplet loss: a deep quadruplet network for person re-identification <https://arxiv.org/abs/1704.01719>`__ for more information.

    It is an extension of `torch.nn.TripletMarginLoss`, where samples
    from two different `negative` (`negative` and `negative2`)
    classes should be pushed further away
    in space than those belonging to the same class (`anchor` and `positive`)

    The loss function for each sample in the mini-batch is:

    .. math::
        L(a, p, n) = \max \{d(a, p) - d(a, n) + {\rm alpha1}, 0\} + \max \{d(a, p) - d(n, n2) + {\rm alpha2}, 0\}

    Arguments
    ---------
    alpha1: float, optional
        Margin of standard `triplet` loss. Default: `1.0`
    alpha2: float, optional
        Margin of second part of loss (pushing negative1 and negative2 samples
        more than positive and anchor). Default: `0.5`
    metric: Callable(torch.Tensor, torch.Tensor) -> torch.Tensor, optional
        Metric used to rate distance between samples. Fully Connected neural
        network with one output and `sigmoid` could be used (as in original paper)
        or anything else adhering to API.
        Default: Euclidean distance.
    weight: Tensor, optional
        Manual rescaling weight, if provided it's repeated to match input
        tensor shape. Default: `None` (no weighting)
    reduction: typing.Callable(torch.Tensor) -> torch.Tensor, optional
        Specifies the reduction to apply to the output.
        If user wants no reduction he should use: `lambda loss: loss`.
        If user wants a summation he should use: `torch.sum`.
        By default, `lambda loss: loss.sum() / loss.shape[0]` is used (mean across examples).


    """

    def __init__(
        self,
        alpha1: float = 1.0,
        alpha2: float = 0.5,
        metric: typing.Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = torch.nn.functional.pairwise_distance,
        weight=None,
        reduction: str = "sum",
    ):
        super().__init__()

        self.alpha1 = alpha1
        self.alpha2 = alpha2
        self.metric = metric

        self.weight = weight
        self.reduction = reduction

    def forward(
        self,
        anchor: torch.Tensor,
        positive: torch.Tensor,
        negative: torch.Tensor,
        negative2: torch.Tensor,
    ) -> torch.Tensor:
        """
        Arguments
        ---------
        anchor: torch.Tensor
            :math:`(N, *)` where :math:`*` means, any number of additional dimensions
            For images usually of shape :math:`(N, C, H, W)`.
        positive: torch.Tensor
            Same as `anchor`
        negative: torch.Tensor
            Same as `anchor`
        negative2: torch.Tensor
            Same as `anchor`

        Returns
        -------
        torch.Tensor
            If :attr:`reduction` is not specified then `mean` across sample is taken.
            Otherwise whatever shape `reduction` returns.
        """
        return functional.loss.quadruplet(
            anchor,
            positive,
            negative,
            negative2,
            self.alpha1,
            self.alpha2,
            self.metric,
            self.weight,
            self.reduction,
        )



"""Perform single step on data and via specific module(s).

.. note::

    **IMPORTANT**: This module is one of core features
    so be sure to understand how it works.
    It is the core and defines how you perform single
    step through the data.

See `Introduction tutorial <https://colab.research.google.com/drive/19oI8RlpDT9JZnkW8BbFzrLL1Wse6wD5G?usp=sharing>`_ for example of `step`.

Usually it looks something along those lines::

    class Step(tt.steps.Step):
        def forward(self, module, batch):
            images, labels = batch
            images, labels = images.to(self.device), labels.to(self.device)

            predictions = module(images)
            loss = self.criterion(predictions, labels)

            return loss, predictions, labels

    step = Step(criterion=torch.nn.BCEWithLogitsLoss, device=torch.device("cuda"))


.. note::

    **IMPORTANT**: You can override `__init__` if you wish to pass
    other arguments.


.. note::

    **IMPORTANT**: You can override `forward` signature to anything you
    desire. Just be sure to pass appropriate data to it (via `iteration` or `epoch`)
    or simple `__call__`.

.. note::

    **IMPORTANT**: `module` is passed from other objects and can be anything.
    In case of GANs in tutorial this is a `Tuple` of `torch.nn.Module`.


"""

import abc
import collections
import dataclasses
import typing

import torch

from . import _base, utils
from .utils import steps as steps_utils


@steps_utils.docstring(
    header="General `step`, usable both in training & evaluation.",
    body="User should override `forward` method.",
)
class Step(_base.Step):
    def __init__(
        self, criterion: typing.Callable, gradient, device=None,
    ):
        super().__init__()

        # Criterion
        self.criterion = criterion
        self.gradient = gradient
        self.device = device

    def __call__(self, *args, **kwargs):
        with torch.set_grad_enabled(self.gradient):
            return super().__call__(*args, **kwargs)

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass


@steps_utils.docstring(
    header="Perform user specified training step with enabled gradient.",
    body="Users should override forward method.",
)
class Train(Step):
    def __init__(
        self, criterion: typing.Callable, device=None,
    ):
        super().__init__(criterion, True, device)

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass


@steps_utils.docstring(
    header="Perform user specified evaluation step with disabled gradient.",
    body="Users should override forward method.",
)
class Eval(Step):
    def __init__(
        self, criterion: typing.Callable, device=None,
    ):
        super().__init__(criterion, False, device)

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass



import torch


def random_split(
    dataset: torch.utils.data.Dataset, *p: float, generator=torch.default_generator
):
    """Randomly split a dataset into non-overlapping new datasets of given proportions.

    Works like `torch.utils.data.random_split` **except** data is splitted on
    [0, 1] proportions instead of length.

    Example::


        train, validation = tt.functional.data.random_split(
            torchvision.datasets.CIFAR10(
                root=".",
                download=True
                transform=torchvision.transforms.ToTensor(),
            ),
            0.8,
            0.2,
        )

    Above would be split dataset into 80% train and 20% validation.

    Arguments
    ---------
    dataset: torch.utils.data.Dataset
        Dataset to be split
    *p: float
        Floating point values in the `[0, 1]`. All of them should sum to `1`
        (if not they will be normalized to `[0, 1]` range). Split dataset
        according to those proportions.
    generator: Generator
        Generator used for the random permutation.

    Returns
    -------
    Tuple[torch.utils.data.Dataset]
        Tuple containing splitted datasets.

    """
    if len(p) < 2:
        raise ValueError(
            "number of proportions should be at least 2, got {}".format(len(p))
        )
    # Normalize to [0, 1] range
    p = torch.tensor(p)
    p = p / torch.sum(p)
    lengths = [int(len(dataset) * split) for split in p[:-1]]
    return torch.utils.data.random_split(
        dataset, lengths + [len(dataset) - sum(lengths)], generator=generator
    )



"""This module allows user to train networks in distributed manner using `horovod`

.. note::

    **IMPORTANT**: This module is experimental and may not be working
    correctly. Use at your own risk and report any issues you find.

.. note::

    **IMPORTANT**: This module needs `horovod` Python package to be visible.
    You can install it with `pip install -U torchtraining[horovod]`.
    Also you should export `CUDA_HOME` variable like this:
    `CUDA_HOME=/opt/cuda pip install -U torchtraining[horovod]` (your path may vary)



See `Horovod documentation <https://github.com/horovod/horovod>`__ for details
about the framework (installation, capabilities etc.).


Example::

    import torchtraining as tt
    import torchtraining.accumulators.horovod as horovod


    class TrainStep(tt.steps.Train):
        def forward(self, module, sample):
            # Dummy step
            images, labels = sample
            return loss


    model = ...
    criterion = ...
    dataset = ...
    optimizer = ...
    writer = ...


    # Accelerate!
    accelerator = tt.accelerators.Horovod(model, optimize.optimizer)

    # Distributed optimization with gradient accumulation
    optimizer = horovod.optimizer(optimizer, module.named_parameters())

    # Special distributed DataLoader
    dataloader = horovod.DataLoader(dataset, batch_size=64)


    step = (
        TrainStep(criterion, device)
        ** tt.pytorch.ZeroGrad()
        ** tt.pytorch.Backward()
        ** tt.pytorch.Optimize(optimizer)
    )
    iteration = (
        ** tt.iterations.TrainIteration(step, model, dataloader)
        ** horovod.AllReduce()
        ** tt.accumulators.Mean()
        ** horovod.OnRank(tt.callbacks.Tensorboard(writer, "Loss"))
    )

Specific `operations` integrated by `torchtraining` below.

"""

import operator
import pathlib
import pickle
import typing

import torch

import horovod.torch as hvd

from .._base import Operation


def _reduction(name):
    mapping = {
        "sum": hvd.Sum,
        "mean": hvd.Average,
    }
    value = mapping.get(name.lower())
    if value is None:
        raise ValueError(
            "reduction can be one of {}, got {}".format(mapping.keys(), name)
        )
    return value


def _compression(name):
    mapping = {
        "none": hvd.compression.NoneCompressor(),
        "fp16": hvd.compression.FP16Compressor(),
    }
    value = mapping.get(name.lower())
    if value is None:
        raise ValueError(
            "compression can be one of {}, got {}".format(mapping.keys(), compression)
        )
    return value


class OnRank(Operation):
    """Run any operation only if it runs in specified process (specified rank).

    Otherwise return unchanged `data`.

    Parameters
    ----------
    operation: tt.Operation
        Operation to run (`callbacks`, `metrics` and whatever else you want).
    rank: int, optional
        Rank (process) on which the operation will be run. Default: `0` (main process)

    Returns
    -------
    data | operation(data)
        If run in specified process, return `operation(data)`. Otherwise forward
        data without changes.

    """

    def __init__(
        self, operation: Operation, rank: int = 0,
    ):
        self.operation = operation
        self.rank = rank

    def forward(self, data: typing.Any):
        """
        Arguments
        ---------
        data: Any
            Input required by `operation`

        """
        if hvd.rank() == self.rank:
            return self.operation(data)
        return data


class DataLoader(torch.utils.data.DataLoader):
    """PyTorch `torch.utils.data.DataLoader` suited for `horovod` integration.

    Works exactly like it's PyTorch counterpart but creates appropriate
    `torch.utils.data.DistributedSampler` under the hood (hence users cannot
    specify `sampler` or `batch_sampler`).

    Arguments:
    ----------
    dataset: torch.utils.data.Dataset
        Dataset from which to load the data.
    batch_size: int, optional
        How many samples per batch to load. Default: ``1``
    shuffle: bool, optional
        Set to ``True`` to have the data reshuffled at every epoch.
        Default: ``False``
    num_workers: int, optional
        How many subprocesses to use for data loading.
        ``0`` means that the data will be loaded in the main process.
        Default: ``0``
    collate_fn Callable, optional
        Merges a list of samples to form a mini-batch of Tensor(s).
        Used when using batched loading from a map-style dataset.
        Default: `None` (default PyTorch collation)
    pin_memory: bool, optional
        If ``True``, the data loader will copy `torch.Tensors`
        into CUDA pinned memory before returning them. Default: `False`
    drop_last: bool, optional
        Set to ``True`` to drop the last incomplete batch,
        if the dataset size is not divisible by the batch size. If ``False`` and
        the size of dataset is not divisible by the batch size, then the last batch
        will be smaller. Default: ``False``
    timeout: Numeric, optional
        If positive, the timeout value for collecting a batch
        from workers. Should always be non-negative.
        Default: ``0``
    worker_init_fn: Callable, optional
        If not ``None``, this will be called on each
        worker subprocess with the worker id (an int in ``[0, num_workers - 1]``) as
        input, after seeding and before data loading.
        Default: ``None``

    """

    def __init__(
        self,
        dataset,
        batch_size=1,
        shuffle=False,
        num_workers=0,
        collate_fn=None,
        pin_memory=False,
        drop_last=False,
        timeout=0,
        worker_init_fn=None,
        multiprocessing_context=None,
        generator=None,
        sampler_seed=0,
    ):
        super().__init__(
            dataset,
            batch_size,
            False,
            torch.utils.data.distributed.DistributedSampler(
                dataset,
                num_replicas=hvd.size(),
                rank=hvd.rank(),
                shuffle=shuffle,
                seed=sampler_seed,
            ),
            None,
            num_workers,
            collate_fn,
            pin_memory,
            drop_last,
            timeout,
            worker_init_fn,
            multiprocessing_context,
            generator,
        )


class AllReduce(Operation):
    """Perform reduction of the input tensor over all the processes.

    If `data` requires gradient you can backpropagate through this operation.

    Parameters
    ----------
    reduction: str, optional
        The reduction operation to use when combining gradients across different
        processes. Can be one of: ["mean", "sum"] being respectively:
        [hvd.mpi_ops.Average, hvd.mpi_ops.Sum].
        Default: "mean"
    compression: str, optional
        Compression algorithm used during allreduce to reduce the amount of
        data sent during the each parameter update step.
        Can be one of "none" or "fp16". Default: "none"
    name: str, optional
        Name of the reduction operator. If not provided it will be generated
        automatically. Default: `None` (automatic generation)

    Returns
    -------
    torch.Tensor
        Tensor with the same shape as `data` averaged (`reduction="mean"`) or
        summed (`reduction="sum"`) across all processes.

    """

    def __init__(self, reduction: str = "mean", compression="none", name=None):
        self.name = name
        self.reduction = _reduction(reduction)
        self.compression = _compression(compression)

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor to be reduced
        """
        return hvd.allreduce(
            data, name=self.name, compression=self.compression, op=self.reduction
        )


class AsyncAllReduce(Operation):
    """Perform asynchronous reduction of the input tensor over all the processes.

    User should pipe this object into `tt.accelerators.horovod.Synchronize()`
    in order to get value.

    Parameters
    ----------
    reduction: str, optional
        The reduction operation to use when combining gradients across different
        processes. Can be one of: ["mean", "sum"] being respectively:
        [hvd.mpi_ops.Average, hvd.mpi_ops.Sum].
        Default: "mean"
    compression: str, optional
        Compression algorithm used during allreduce to reduce the amount of
        data sent during the each parameter update step.
        Can be one of "none" or "fp16". Default: "none"
    name: str, optional
        Name of the reduction operator. If not provided it will be generated
        automatically. Default: `None` (automatic generation)


    Returns
    -------
    Handle
        Handle to be used with `tt.accelerators.horovod.Synchronize()`

    """

    def __init__(self, reduction: str = "mean", compression="none", name=None):
        self.name = name
        self.reduction = _reduction(reduction)

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor to be reduced across all processes.
        """
        return hvd.allreduce_async(data, name=self.name, op=self.reduction)


class AllGather(Operation):
    """Concatenate input tensors from all processes.

    Tensor after concatenation will be available to all processes.
    Concatenation is done over `0` th dimension, so it's the only dimension
    in which `torch.Tensor` on different processes is allowed to be different.

    If `data` requires gradient you can backpropagate through this operation.

    Parameters
    ----------
    name: str, optional
        Name of the reduction operator. If not provided it will be generated
        automatically. Default: `None` (automatic generation)

    Returns
    -------
    torch.Tensor
        Tensor with the same shape as `data` except `0` dimension (which will be larger
        as it's concatenation of data from all processes).

    """

    def __init__(self, name: str = None):
        self.name = name

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor to be gathered across all processes.
        """
        return hvd.allgather(data, name=self.name)


class AsyncAllGather(Operation):
    """Asynchronously concatenate input tensors from all processes.

    Tensor after concatenation will be available to all processes.
    Concatenation is done over `0`th dimension, so it's the only dimension
    in which `torch.Tensor` on different processes is allowed to be different.

    Parameters
    ----------
    name: str, optional
        Name of the reduction operator. If not provided it will be generated
        automatically. Default: `None` (automatic generation)

    Returns
    -------
    Handle
        Handle to be used with `tt.accelerators.horovod.Synchronize()`

    """

    def __init__(self, name=None):
        self.name = name

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor to be gathered across all processes.
        """
        return hvd.allgather_async(data, name=self.name,)


class Broadcast(Operation):
    """Broadcast tensor from `rank` process to all other processes.

    If `data` requires gradient you can backpropagate through this operation.

    Parameters
    ----------
    rank: int, optional
        Rank of the process from which `data` will be distributed to other processes.
    name: str, optional
        Name of the reduction operator. If not provided it will be generated
        automatically. Default: `None` (automatic generation)

    Returns
    -------
    torch.Tensor
        Tensor with the same shape as `data` with broadcasted values.

    """

    def __init__(self, rank: int = 0, name=None):
        self.rank = rank
        self.name = name

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor to be broadcasted across all processes.
        """
        return hvd.broadcast(data, self.rank, name=self.name)


class AsyncBroadcast(Operation):
    """Asynchronously broadcast tensor from `rank` process to all other processes.

    Parameters
    ----------
    rank: int, optional
        Rank of the process from which `data` will be distributed to other processes.
    name: str, optional
        Name of the reduction operator. If not provided it will be generated
        automatically. Default: `None` (automatic generation)

    Returns
    -------
    Handle
        Handle to be used with `tt.accelerators.horovod.Synchronize()`

    """

    def __init__(self, rank: int = 0, name=None):
        self.rank = rank
        self.name = name

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor to be broadcasted across all processes.
        """
        return hvd.async_broadcast(data, self.rank, name=self.name)


class Synchronize(Operation):
    """Asynchronously broadcast tensor from `rank` process to all other processes.

    Returns
    -------
    torch.Tensor
        Value of the previous asynchronous operation after synchronization.
        Whatever it should return.

    """

    def forward(self, handle):
        """
        Arguments
        ---------
        handle: Handle
            Handle returned by an `AsyncAllReduce`, `AsyncAllGather`or
            `AsyncBroadcast` which will be used to retrieve `torch.Tensor`.
        """
        return hvd.synchronize(handle)


def optimizer(
    optimizer,
    named_parameters,
    reduction: str = "sum",
    compression: str = "none",
    accumulate: int = 1,
    rank: int = 0,
):
    """Create Horovod compatible optimizer.

    State of optimizer will be distributed on specified `rank`.
    Should be used after `torchtraining.accelerators.Horovod` object was created.

    Parameters
    ----------
    optimizer: torch.optim.Optimizer
        Instance of optimizer-like object with interface aligned with
        `torch.optim.Optimizer`.
    named_parameters: torch.nn.parameter.Parameter
        A mapping between parameter names and values. Used for naming of allreduce operations.
        Typically just `model.named_parameters()`.
    reduction: str, optional
        The reduction operation to use when combining gradients across different
        processes. Can be one of: ["mean", "sum"] being respectively:
        [hvd.mpi_ops.Average, hvd.mpi_ops.Sum].
        Default: "mean"
    compression: str, optional
        Compression algorithm used during allreduce to reduce the amount of
        data sent during the each parameter update step.
        Can be one of "none" or "fp16". Default: "none"
    accumulate: int, optional
        Divide loss by ``accumulate`` if gradient accumulation is used.
        This approach averages gradient from multiple batches.
        Default: `1` (no accumulation)
    rank: int, optional
        Rank from which optimizer's state will be broadcasted.
        Default: `0`

    Returns
    -------
    horovod.torch.DistributedOptimizer
        Instance of optimizer but distributed across workers.

    """

    optimizer = hvd.DistributedOptimizer(
        optimizer,
        named_parameters,
        _compression(compression),
        backward_passes_per_step=accumulate,
        op=_reduction(reduction),
    )

    hvd.broadcast_optimizer_state(optimizer, root_rank=rank)
    return optimizer


def load(f, rank: int = 0, map_location=None, pickle_module=pickle, **pickle_load_args):
    """Load object saved with `torch.save` in a single process and distribute to all other processes.

    Useful when loading saved `torch.nn.Module` (or other `torch` objects like `optimizer`),
    which is saved on a single machine.

    It can be easily distributed to other processes this way.

    If you wish to `torch.save` on a single process you can create an object
    like this::

        save = tt.accelerators.horovod.OnRank(torch.save)
        save(your_module)

    Arguments
    ---------
    f: file-like
        A file-like object (has to implement :meth:`read`, :meth`readline`, :meth`tell`, and :meth`seek`)
        or a string or `os.PathLike` object containing a file name.
    rank: int, optional
        Process rank on which data will be loaded.
    map_location: Callable | `torch.device` | string | dict, optional
        Specifies how to remap storage locations. Default: `None`
    pickle_module: module, optional
        Module used for unpickling metadata and objects,
        (has to match the :attr:`pickle_module` used to serialize file).
        Default: `pickle`
    **pickle_load_args
        optional keyword arguments passed over to :func:`pickle_module.load`
        and :func:`pickle_module.Unpickler`, e.g., :attr:`errors=...`.

    Returns
    -------
    torch.Tensor | torch.nn.Module | Any
        Anything you saved with `torch.save` really

    """
    data = None
    if hvd.rank() == rank:
        data = torch.load(f, map_location, pickle_module, **pickle_load_args)
    return hvd.broadcast(data, rank)



"""Special type of callbacks focused on `tensorboard` integration.

.. note::
    **IMPORTANT**: Users need `tensorboard` package installed for this
    module to exist.
    You can install it via `pip install torchtraining[tensorboard]`
    (additional libraries for `Image`, `Images`, `Video`, `Figure`)
    will also be installed)
    or install `tensorboard` directly via `pip install -U tensorboard`
    or a-like command (in this case not all functions may be available,
    see PyTorch's `torch.utils.tensorboard.SummaryWriter` docs for
    exact packages needed for each functionality).


Example::

    # Assume iteration was defined and loss is 0th element of step
    iteration = (
        tt.iterations.Iteration(...)
        ** tt.Select(loss=0)
        ** tt.device.CPU()
        ** tt.accumulators.Mean()
        ** tt.callbacks.tensorboard.Scalar(writer, "Network/Loss")
    )


"""

import typing

import loguru

from .. import _base
from ..utils import general as utils


def _docs(body=None):
    def wrapper(klass):
        function = klass.__name__.lower()
        docstring = """Log {function} to Tensorboard's summary.

User should specify single `writer` instance to all `torchtraining.callbacks.tensorboard`
objects used for training.

See `torch.utils.tensorboard.writer.add_{function}` for more details.

Can be used similarly to `torchtraining.callbacks.Logger`

        """.format(
            function=function,
        )
        if body is not None:
            docstring += body

        docstring += """
Parameters
----------
writer: torch.utils.tensorboard.SummaryWriter
    Writer responsible for logging values.
name: str
    Name (tag) under which values will be logged into Tensorboard.
    Can be "/" separated to group values together, e.g. "Classifier/Loss"
    and "Classifier/Accuracy"
flush: int
    Flushes the event file to disk after `flush` steps.
    Call this method to make sure that all pending events have been written to disk.
log : str | int, optional
    Severity level for logging object's actions.
    Available levels of logging:
        * NONE          0
        * TRACE 	5
        * DEBUG 	10
        * INFO 	        20
        * SUCCESS 	25
        * WARNING 	30
        * ERROR 	40
        * CRITICAL 	50
    Default: `NONE` (no logging, `0` priority)
*args
    Variable length arguments passed to `add_{function}` call.
**kwargs
    Keyword variable length arguments passed to `add_{function}` call.

        """.format(
            function=function,
        )
        klass.__doc__ = docstring
        return klass

    return wrapper


class _Tensorboard(_base.Operation):
    def __init__(
        self,
        writer,
        name: str,
        flush: int = None,
        log: typing.Union[str, int] = "NONE",
        *args,
        **kwargs
    ):
        super().__init__()

        self.writer = writer
        self.name: str = name
        self.flush = flush
        self.log = log
        self.args = args
        self.kwargs = kwargs
        self._step = -1

    def forward(self, *data):
        """
        Arguments
        ---------
        data:
            Tensor (or a-like) to be logged into Tensorboard

        """
        self._step += 1
        class_name = type(self).__name__
        getattr(self.writer, "add_{}".format(class_name.lower()))(
            self.name, *data, self._step, *self.args, **self.kwargs
        )
        loguru.logger.log(
            self.log, "{} added to Tensorboard.".format(class_name),
        )
        if (self.flush is not None) and (self.flush % self._step == 0):
            loguru.logger.log(self.log, "Events flushed to disk.")
            self.writer.flush()

        return data


@_docs()
class Scalar(_Tensorboard):
    pass


@_docs()
class Scalars(_Tensorboard):
    pass


@_docs()
class Histogram(_Tensorboard):
    pass


if utils.modules_exist("PIL"):

    @_docs()
    class Image(_Tensorboard):
        pass

    @_docs()
    class Images(_Tensorboard):
        pass


if utils.modules_exist("matplotlib"):

    @_docs(body=r"Note that this requires the `matplotlib` package.",)
    class Figure(_Tensorboard):
        pass


if utils.modules_exist("moviepy"):

    @_docs(body=r"Note that this requires the `moviepy` package.",)
    class Video(_Tensorboard):
        pass


@_docs()
class Audio(_Tensorboard):
    pass


@_docs()
class Text(_Tensorboard):
    pass


@_docs()
class Mesh(_Tensorboard):
    pass


@_docs()
class PRCurve(_Tensorboard):
    def forward(self, data):
        predictions, labels = data
        return super().forward(labels, predictions)



import itertools

import torch

import pytest
import torchtraining as tt


@pytest.mark.parametrize(
    "loss,outputs",
    list(
        itertools.product(
            (
                tt.loss.BinaryFocal(gamma=0.2,),
                tt.loss.BinaryFocal(
                    gamma=0.5,
                    weight=1.2,
                    pos_weight=torch.tensor(2),
                    reduction=torch.sum,
                ),
                tt.loss.SmoothBinaryCrossEntropy(alpha=0.1),
                tt.loss.SmoothBinaryCrossEntropy(alpha=0.0),
                tt.loss.SmoothBinaryCrossEntropy(
                    alpha=0.4,
                    weight=1.5,
                    pos_weight=torch.tensor(2),
                    reduction=torch.sum,
                ),
            ),
            (torch.randn(4, 3, 2, 2), torch.randn(4, 8), torch.randn(1)),
        )
    ),
)
def test_binary_smoke(loss, outputs):
    targets = torch.randn_like(outputs) < 0.5
    loss(outputs, targets)


@pytest.mark.parametrize(
    "loss,outputs",
    list(
        itertools.product(
            (
                tt.loss.MulticlassFocal(gamma=0.2,),
                tt.loss.MulticlassFocal(
                    gamma=0.5, ignore_index=1, reduction=torch.sum,
                ),
                tt.loss.SmoothCrossEntropy(alpha=0.1),
                tt.loss.SmoothCrossEntropy(alpha=0.0),
                tt.loss.SmoothCrossEntropy(
                    alpha=0.4, ignore_index=0, reduction=torch.sum,
                ),
            ),
            (torch.randn(4, 7, 3), torch.randn(4, 7), torch.randn(4, 8, 3, 2)),
        )
    ),
)
def test_multiclass_smoke(loss, outputs):
    targets = torch.argmax(outputs, dim=1)
    loss(outputs, targets)



import pytest
import torch
import torchtraining as tt


class Step(tt.steps.Train):
    def forward(self, module, sample):
        data, target = sample
        predictions = module(data).squeeze()
        loss = self.criterion(predictions, target)
        return loss, predictions, target


def test_step():
    module = torch.nn.Linear(10, 1)
    optimizer = torch.optim.Adam(module.parameters())
    step = (
        Step(torch.nn.MSELoss())
        ** tt.Select(loss=0)
        ** tt.pytorch.ZeroGrad(optimizer)
        ** tt.pytorch.Backward(accumulate=2)
        ** tt.pytorch.Optimize(optimizer, accumulate=2)
        ** tt.pytorch.Detach()
        ** tt.callbacks.Log("Loss")
    )

    step = Step(torch.nn.MSELoss()) ** tt.Select(loss=0)

    step ** tt.Select(
        predictions=1, target=2
    ) ** tt.metrics.regression.MaxError() ** tt.callbacks.Log(name="Max Error")

    for _ in range(5):
        step(module, (torch.randn(8, 10), torch.randn(8)))



import torch

from .. import _base, functional
from . import utils

###############################################################################
#
#                               NO REDUCTION
#
###############################################################################


@utils.distance.docstring(header="Cosine distance between `outputs` and `targets`")
class Cosine(_base.Operation):
    """Returns cosine distance between :math:`x_1` and :math:`x_2`, computed along dim.

    It is equal to :math:`1 - S`, where :math:`S` is cosine similarity.

    Parameters
    ----------
    dim: int, optional
        Dimension where cosine similarity is computed. Default: 1
        Degree of `norm`. Default: `2`
    eps: float, optional
        Epsilon to avoid division by zero. Default: `1e-08`
    reduction: Callable(torch.Tensor) -> Any, optional
        One argument callable getting `torch.Tensor` and returning argument
        after specified reduction.
        Default: `torch.mean` (mean across batch, user can use `torchtraining.savers.Mean`
        to get mean across iterations/epochs).


    """

    def __init__(
        self, dim: int = 1, eps: float = 1e-08, reduction=torch.mean,
    ):
        super().__init__()
        self.dim = dim
        self.eps = eps
        self.reduction = reduction

    def forward(self, data):
        """
        Arguments
        ---------
        data: Tuple[torch.Tensor, torch.Tensor]
            Tuple containing `outputs` from neural network and regression `targets`.
            `outputs` should be of shape :math:`(N, F)`, where :math:`N` is the number of samples,
            :math:`F` is the number of features.
            Should contain `floating` point values.
            `targets` should be in the same shape `outputs` and be of `float` data type as well.

        Returns
        -------
        torch.Tensor
            If `reduction` is left as default {} is taken and single value returned.
            Otherwise whatever `reduction` returns.
        """
        return functional.metrics.distance.cosine(
            *data, self.dim, self.eps, self.reduction
        )


###############################################################################
#
#                              OTHER METRICS
#
###############################################################################


class Pairwise(_base.Operation):
    """Computes the batchwise pairwise distance between vectors :math:`v_1`, :math:`v_2` using specified norm.

    Parameters
    ----------
    p: float, optional
        Degree of `norm`. Default: `2`
    eps: float, optional
        Epsilon to avoid division by zero. Default: `1e-06`
    reduction: Callable(torch.Tensor) -> Any, optional
        One argument callable getting `torch.Tensor` and returning argument
        after specified reduction.
        Default: `torch.mean` (mean across batch, user can use `torchtraining.savers.Mean`
        to get mean across iterations/epochs).
    """

    def __init__(self, p: float = 2.0, eps: float = 1e-06, reduction=torch.mean):
        super().__init__()
        self.p = p
        self.eps = eps
        self.reduction = reduction

    def forward(self, data):
        """
        Arguments
        ---------
        data: Tuple[torch.Tensor, torch.Tensor]
            Tuple containing `outputs` from neural network and regression `targets`.
            `outputs` should be of shape :math:`(N, F)`, where :math:`N` is the number of samples,
            :math:`F` is the number of features.
            Should contain `floating` point values.
            `targets` should be in the same shape `outputs` and be of `float` data type as well.

        Returns
        -------
        torch.Tensor
            If `reduction` is left as default {} is taken and single value returned.
            Otherwise whatever `reduction` returns.
        """
        return functional.metrics.distance.pairwise(
            *data, self.p, self.eps, self.reduction,
        )



import torch

from .. import _base, functional
from . import utils

###############################################################################
#
#                           COMMON BASE CLASSES
#
###############################################################################


class _MeanReduction(_base.Operation):
    def __init__(self, reduction=torch.mean):
        super().__init__()

        self.reduction = reduction


class _SumReduction(_base.Operation):
    def __init__(self, reduction=torch.sum):
        super().__init__()

        self.reduction = reduction


###############################################################################
#
#                          CONCRETE IMPLEMENTATIONS
#
###############################################################################


###############################################################################
#
#                           MEAN DEFAULT REDUCTION
#
###############################################################################


@utils.regression.docstring(
    header="""Absolute error between `outputs` and `targets`.""", reduction="mean",
)
class AbsoluteError(_MeanReduction):
    def forward(self, data):
        return functional.metrics.regression.absolute_error(*data, self.reduction)


@utils.regression.docstring(
    header="""Squared error between `outputs` and `targets`.""", reduction="mean",
)
class SquaredError(_MeanReduction):
    def forward(self, data):
        return functional.metrics.regression.squared_error(*data, self.reduction)


@utils.regression.docstring(
    header="""Squared log error between `outputs` and `targets`.""", reduction="mean",
)
class SquaredLogError(_MeanReduction):
    def forward(self, data):
        return functional.metrics.regression.squared_log_error(*data, self.reduction)


###############################################################################
#
#                           SUM DEFAULT REDUCTION
#
###############################################################################


@utils.regression.docstring(
    header="""Regression of squares between `outputs` and `targets`.""",
    reduction="sum",
)
class RegressionOfSquares(_SumReduction):
    def forward(self, data):
        return functional.metrics.regression.regression_of_squares(
            *data, self.reduction
        )


@utils.regression.docstring(
    header="""Square of residuals between `outputs` and `targets`.""", reduction="sum",
)
class SquaresOfResiduals(_SumReduction):
    def forward(self, data):
        return functional.metrics.regression.squares_of_residuals(*data, self.reduction)


###############################################################################
#
#                               NO REDUCTION
#
###############################################################################


@utils.regression.docstring(header="""R2 score between `outputs` and `targets`.""",)
class R2(_base.Operation):
    def forward(self, data):
        return functional.metrics.regression.r2(*data)


@utils.regression.docstring(
    header="""Maximum error between `outputs` and `targets`.""",
)
class MaxError(_base.Operation):
    def forward(self, data):
        return functional.metrics.regression.max_error(*data)


###############################################################################
#
#                               OTHER METRICS
#
###############################################################################


class TotalOfSquares(_SumReduction):
    """Total of squares of single `tensor`.

    Parameters
    ----------
    reduction: Callable, optional
        One argument callable getting `torch.Tensor` and returning `torch.Tensor`.
        Default: `torch.sum` (sum of all elements, user can use `torchtraining.savers.Sum`
        to get sum across iterations/epochs).

    """

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor containing `float` data of any shape. Usually `targets`.

        Returns
        -------
        torch.Tensor
            If `reduction` is left as default {} is taken and single value returned.
            Otherwise whatever `reduction` returns.
        """
        return functional.metrics.regression.total_of_squares(data, self.reduction)


class AdjustedR2(_base.Operation):
    """Adjusted R2 score between `outputs` and `targets`.

    Parameters
    ----------
    p: int
        Number of explanatory terms in model.

    """

    def __init__(self, p: int):
        super().__init__()

        self.p = p

    def forward(self, data):
        """
        Arguments
        ---------
        data: Tuple[torch.Tensor, torch.Tensor]
            Tuple containing `outputs` from neural network and regression `targets`.
            `outputs` should be of shape :math:`(N, *)`, where :math:`N` is the number of samples.
            Should contain `floating` point values.
            `targets` should be in the same shape `outputs` and be of `float` data type as well.

        Returns
        -------
        torch.Tensor
            Scalar `tensor`

        """
        return functional.metrics.regression.adjusted_r2(*data, self.p)



import typing

import torch
import torchtraining as tt
import torchvision


def get(batch_size: int) -> typing.Dict:
    train, validation = tt.functional.data.random_split(
        torchvision.datasets.FakeData(
            size=64,
            image_size=(3, 28, 28),
            transform=torchvision.transforms.ToTensor(),
        ),
        0.8,
        0.2,
    )
    test = torchvision.datasets.FakeData(
        size=32, image_size=(3, 28, 28), transform=torchvision.transforms.ToTensor()
    )
    return {
        "train": torch.utils.data.DataLoader(train, batch_size, shuffle=True),
        "validation": torch.utils.data.DataLoader(validation, batch_size),
        "test": torch.utils.data.DataLoader(test, batch_size),
    }



import itertools

import numpy as np
import torch
from sklearn.metrics import (max_error, mean_absolute_error,
                             mean_squared_error, mean_squared_log_error,
                             r2_score)

import pytest
import torchtraining as tt

# @pytest.mark.parametrize(
#     "metrics,data",
#     list(
#         itertools.product(
#             (
#                 (tt.metrics.regression.AbsoluteError(), mean_absolute_error),
#                 (tt.metrics.regression.SquaredError(), mean_squared_error),
#                 (tt.metrics.regression.SquaredLogError(), mean_squared_log_error),
#                 (tt.metrics.regression.R2(), r2_score),
#                 (tt.metrics.regression.MaxError(), max_error),
#             ),
#             list(
#                 itertools.permutations(
#                     (
#                         torch.randn(10),
#                         torch.ones(10),
#                         torch.zeros(10),
#                         torch.randn(10),
#                     ),
#                     r=2,
#                 )
#             ),
#         ),
#     ),
# )
# def test_regression(metrics, data):
#     package, scikit = metrics
#     outputs, targets = data
#     our = package(data).numpy()
#     their = scikit(targets.numpy(), outputs.numpy())
#     assert np.isclose(our, their)


@pytest.mark.parametrize(
    "metric,data",
    list(
        itertools.product(
            (
                tt.metrics.regression.AbsoluteError(),
                tt.metrics.regression.SquaredError(),
                tt.metrics.regression.SquaredLogError(),
                tt.metrics.regression.R2(),
                tt.metrics.regression.MaxError(),
                tt.metrics.regression.RegressionOfSquares(),
                tt.metrics.regression.SquaresOfResiduals(),
                tt.metrics.regression.AdjustedR2(p=6),
            ),
            (
                (torch.randn(8, 1), torch.randn(8, 1)),
                (torch.randn(2, 2, 2), torch.randn(2, 2, 2)),
            ),
        )
    ),
)
def test_smoke(metric, data):
    metric(data)


@pytest.mark.parametrize(
    "metric,data",
    [
        (tt.metrics.regression.TotalOfSquares(), torch.randn(4, 2, 3),),
        (tt.metrics.regression.TotalOfSquares(), torch.randn(4),),
    ],
)
def test_specific_smoke(metric, data):
    metric(data)



# import torch

# import pytest
# import torchtraining as tt


# @pytest.mark.parametrize("inputs", [range(5), range(0), range(20)])
# def test_sum(inputs):
#     accumulator = tt.accumulators.Sum()
#     for i in inputs:
#         accumulator(i)
#     assert accumulator.calculate() == sum(inputs)


# @pytest.mark.parametrize("inputs", [range(5), range(1), range(20)])
# def test_mean(inputs):
#     accumulator = tt.accumulators.Mean()
#     for i in inputs:
#         accumulator(i)
#     assert accumulator.calculate() == sum(inputs) / len(inputs)


# @pytest.mark.parametrize("inputs", [range(5), range(0), range(20)])
# def test_list(inputs):
#     accumulator = tt.accumulators.List()
#     for i in inputs:
#         accumulator(i)
#     assert accumulator.calculate() == list(inputs)



import itertools

import numpy as np
import torch

import pytest
import torchtraining as tt


@pytest.mark.parametrize(
    "metric,data",
    list(
        itertools.product(
            (tt.metrics.distance.Cosine(), tt.metrics.distance.Pairwise()),
            list(
                itertools.permutations(
                    (torch.randn(4, 10), torch.ones(4, 10), torch.zeros(4, 10),), r=2,
                )
            ),
        ),
    ),
)
def test_smoke(metric, data):
    metric(data)



"""Integrate `torchtraining` with `comet.ml <https://www.comet.ml/site/>`__ experiment management tool.

.. note::

    **IMPORTANT**: This module is experimental and may not be working
    correctly. Use at your own risk and report any issues you find.

.. note::

    **IMPORTANT**: This module needs `comet-ml` Python package to be available.
    You can install it with `pip install -U torchtraining[neptune]`

Usage is similar to `torchtraining.callbacks.Tensorboard`, except creating `Experiment`
instead of `torch.utils.tensorboard.SummaryWriter`.

Example::

    import torchtraining as tt
    import torchtraining.callbacks.comet as comet


    class TrainStep(tt.steps.Train):
        def forward(self, module, sample):
            images, labels = sample
            ...
            return loss, images


    project = comet.Experiment()
    step = TrainStep(criterion, device)

    # You have to split `tensor` as only single image can be logged
    step ** tt.Select(1) ** tt.OnSplittedTensor(comet.Image(experiment))
    step ** tt.Select(0) ** tt.cast.Item() ** comet.Scalar(experiment)


"""

import typing

import comet_ml

from .. import _base

Experiment = comet_ml.Experiment
ExistingExperiment = comet_ml.ExistingExperiment
OfflineExperiment = comet_ml.OfflineExperiment


class Clean(_base.Operation):
    """Clean experiment loggers

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment

    Returns
    -------
    Any
        Data passed initially to the operation.

    """

    def __init__(self, experiment):
        super().__init__()
        self.experiment = experiment

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything, will be forwarded
        """
        self.experiment.clean()
        return data


class Asset(_base.Operation):
    """Logs the Asset passed during call.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    name: str, optional
        A custom file name to be displayed. If not provided the `filename`
        from the file_data argument will be used.
    overwrite: bool, optional
        If True will overwrite all existing assets with the same name.
        Default: `False`
    copy_to_tmp: bool, optional
        If file_data is a file-like object, then this flag determines if the file is
        first copied to a temporary file before upload. If `copy_to_tmp`
        is False, then it is sent directly to the cloud.
        Default: `True`
    step: int, optional
        Used to associate the asset to a specific step. Default: `None`
    metadata: typing.Dict, optional
        Optional. Some additional data to attach to the the asset data.
        Must be a JSON-encodable dict. Default: `None`

    Returns
    -------
    str | File-like
        Data passed initially to the operation.

    """

    def __init__(
        self,
        experiment,
        name=None,
        overwrite=False,
        copy_to_tmp=True,
        step=None,
        metadata=None,
    ):
        super().__init__()
        self.experiment = experiment
        self.name = name
        self.overwrite = overwrite
        self.copy_to_tmp = copy_to_tmp
        self.step = step
        self.metadata = metadata

    def forward(self, data):
        """
        Arguments
        ---------
        data: str | File-like
            Either the file path of the file you want to log, or a file-like asset.
        """
        self.experiment.log_asset(
            data, self.name, self.overwrite, self.copy_to_tmp, self.step, self.metadata,
        )
        return data


class AssetData(_base.Operation):
    """Log given data (`str`, `binary` or `JSON`).

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    name: str, optional
        A custom file name to be displayed. If not provided the `filename`
        from the file_data argument will be used.
    overwrite: bool, optional
        If True will overwrite all existing assets with the same name.
        Default: `False`
    step: int, optional
        Used to associate the asset to a specific step. Default: `None`
    metadata: typing.Dict, optional
        Optional. Some additional data to attach to the the asset data.
        Must be a JSON-encodable dict. Default: `None`


    Returns
    -------
    str | File-like
        Data passed initially to the operation.

    """

    def __init__(
        self, experiment, name=None, overwrite=False, step=None, metadata=None,
    ):
        super().__init__()
        self.experiment = experiment
        self.name = name
        self.overwrite = overwrite
        self.step = step
        self.metadata = metadata

    def forward(self, data):
        """
        Arguments
        ---------
        data: str | JSON
            Data to log
        """
        self.experiment.log_asset_data(
            data, self.name, self.overwrite, self.step, self.metadata,
        )
        return data


class AssetFolder(_base.Operation):
    """Logs all the files located in the given folder as assets.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    step: int, optional
        Used to associate the asset to a specific step. Default: `None`
    log_file_name: bool, optional
        If True, log the file path with each file. Default: `False`
    recursive: bool, optional
        If `true` recurse folder and save file names.
        Default: `False`

    Returns
    -------
    folder: str
        Data passed initially to the operation.

    """

    def __init__(
        self, experiment, step=None, log_file_name=False, recursive=False,
    ):
        super().__init__()
        self.experiment = experiment
        self.step = step
        self.log_file_name = log_file_name
        self.recursive = recursive

    def forward(self, data):
        """
        Arguments
        ---------
        folder: str
            Path to the folder to be logged.

        """
        self.experiment.log_asset_folder(
            data, self.step, self.log_file_name, self.recursive
        )
        return data


class Audio(_base.Operation):
    """Logs the audio Asset determined by audio data.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    sample_rate: int, optional
        The sampling rate given to scipy.io.wavfile.write for creating the wav file.
    name: str, optional
        A custom file name to be displayed. If not provided the `filename`
        from the file_data argument will be used.
    overwrite: bool, optional
        If True will overwrite all existing assets with the same name.
        Default: `False`
    copy_to_tmp: bool, optional
        If file_data is a file-like object, then this flag determines if the file is
        first copied to a temporary file before upload. If `copy_to_tmp`
        is False, then it is sent directly to the cloud.
        Default: `True`
    step: int, optional
        Used to associate the asset to a specific step. Default: `None`
    metadata: typing.Dict, optional
        Optional. Some additional data to attach to the the asset data.
        Must be a JSON-encodable dict. Default: `None`

    Returns
    -------
    data: str | np.array
        Data passed initially to the operation.

    """

    def __init__(
        self,
        experiment,
        sample_rate=None,
        name=None,
        overwrite=False,
        copy_to_tmp=True,
        step=None,
        metadata=None,
    ):
        super().__init__()
        self.experiment = experiment
        self.sample_rate = sample_rate
        self.name = name
        self.overwrite = overwrite
        self.copy_to_tmp = copy_to_tmp
        self.step = step
        self.metadata = metadata

    def forward(self, data):
        """
        Arguments
        ---------
        data: str | np.array
            Either the file path of the file you want to log, or a numpy array given to
            `scipy.io.wavfile.write` for wav conversion.

        """
        self.experiment.log_audio(
            data,
            self.sample_rate,
            self.name,
            self.metadata,
            self.overwrite,
            self.copy_to_tmp,
            self.step,
        )
        return data


class ConfusionMatrix(_base.Operation):
    """Logs confusion matrix.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    title: str, optional
        A custom name to be displayed. By default, it is "Confusion Matrix".
    row_label: str, optional
        Label for rows. By default, it is "Actual Category".
    column_label: str, optional
        Label for columns. By default, it is "Predicted Category".
    max_example_per_cell: int, optional
        Maximum number of examples per cell. By default, it is 25.
    max_categories: int, optional
        Max number of columns and rows to use. By default, it is 25.
    winner_function: Callable(List) -> List, optional
        A function that takes in an entire list of rows of patterns,
        and returns the winning category for each row. By default, it is argmax.
    index_to_example_function: Callable, optional
        A function that takes an index and returns either a number, a string, a URL, or a
        {"sample": str, "assetId": str} dictionary.
        See below for more info.
        By default, the function returns a number representing the index of the example.
    cache: bool, optional
        Should the results of index_to_example_function be cached and reused?
        By default, cache is `True`.
    selected: List, optional
        None, or list of selected category indices.
        These are the rows/columns that will be shown. By default, select is None.
        If the number of categories is greater than max_categories, and selected is not provided,
        then selected will be computed automatically by selecting the most confused categories.
    kwargs: optional
        any extra keywords and their values will be passed onto the index_to_example_function.

    Returns
    -------
    data: List[List[double]]
        Data passed initially to the operation.

    """

    def __init__(
        self,
        experiment,
        title="Confusion Matrix",
        row_label="Actual Category",
        column_label="Predicted Category",
        max_examples_per_cell=25,
        max_categories=25,
        winner_function=None,
        index_to_example_function=None,
        cache=True,
        file_name="confusion-matrix.json",
        overwrite=False,
        step=None,
        **kwargs
    ):
        super().__init__()
        self.experiment = experiment
        self.title = title
        self.row_label = row_label
        self.column_label = column_label
        self.max_examples_per_cell = max_examples_per_cell
        self.max_categories = max_categories
        self.winner_function = winner_function
        self.index_to_example_function = index_to_example_function
        self.cache = cache
        self.file_name = file_name
        self.overwrite = overwrite
        self.step = step
        self.kwargs = kwargs

    def forward(self, data):
        """
        Arguments
        ---------
        data: List[List[double]]
            Matrix-like list contianing `confusion` matrix.

        """
        self.experiment.log_confusion_matrix(
            None,
            None,
            data,
            None,
            self.title,
            self.row_label,
            self.column_label,
            self.max_examples_per_cell,
            self.max_categories,
            self.winner_function,
            self.index_to_example_function,
            self.cache,
            self.file_name,
            self.overwrite,
            self.step,
            **self.kwargs,
        )
        return data


class Curve(_base.Operation):
    """Log timeseries data.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    name: str
        Name of data
    overwrite: bool, optional
        If True overwrite previous log. Default: `False`
    step: int, optional
        Step value. Default: `None`

    Returns
    -------
    data: Tuple(List[Number], List[Number])
        Data passed initially to operation

    """

    def __init__(
        self, experiment, name=None, overwrite=False, step=None,
    ):
        super().__init__()
        self.experiment = experiment
        self.name = name
        self.overwrite = overwrite
        self.step = step

    def forward(self, data):
        """
        Arguments
        ---------
        data: Tuple(List[Number], List[Number])
            Either the file path of the file you want to log, or a numpy array given to
            `scipy.io.wavfile.write` for wav conversion.
        """
        self.experiment.log_curve(
            self.name, *data, self.overwrite, self.step,
        )
        return data


class Embedding(_base.Operation):
    """Log a multi-dimensional dataset and metadata for viewing with Comet's Embedding Projector.

    This feature is currently deemed experimental.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    image_data: List[Array] | Images, optional
        List of arrays or Images
    image_size: int, optional (required if image_data is given)
        The size of each image
    image_preprocess_function: Callable, optional
        If image_data is an array, apply this function to each element first
    image_transparent_color: Tuple, optional
        (red, green, blue) tuple
    image_background_color_function: Callable(int) -> Tuple(red, green, blue), optional
        A function that takes an index, and returns a (red, green, blue) color tuple
    title: str, optional
        Name of tensor
    template_filename: str, optional
        name of template JSON file

    Returns
    -------
    data: Tuple(torch.Tensor, torch.Tensor)
        Tensors to visualize in 3D and labels for each tensor.

    """

    def __init__(
        self,
        experiment,
        image_data=None,
        image_size=None,
        image_preprocess_function=None,
        image_transparent_color=None,
        image_background_color_function=None,
        title="Comet Embedding",
        template_filename="template_projector_config.json",
        group=None,
    ):

        super().__init__()
        self.experiment = experiment

        self.image_data = image_data
        self.image_size = image_size
        self.image_preprocess_function = image_preprocess_function
        self.image_transparent_color = image_transparent_color
        self.image_background_color_function = image_background_color_function
        self.title = title
        self.template_filename = template_filename
        self.group = group

    def forward(self, data):
        """
        Arguments
        ---------
        data: Tuple(torch.Tensor, torch.Tensor)
            Tensors to visualize in 3D and labels for each tensor.
        """
        self.experiment.log_embedding(
            *data,
            self.image_data,
            self.image_size,
            self.image_preprocess_function,
            self.image_transparent_color,
            self.image_background_color_function,
            self.title,
            self.template_filename,
            self.group,
        )
        return data


class Figure(_base.Operation):
    """Logs the global Pyplot figure or the passed one and upload its svg version to the backend.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    figure_name: str, optional
        Name of the figure
    overwrite: bool, optional
        If another figure with the same name exists, it will be overwritten if overwrite is set to True.
        Default: `False`
    step: int, optional
        Used to associate figure to a specific step.

    Returns
    -------
    data: figure, optional.
        Data passed initially to operation.

    """

    def __init__(self, experiment, figure_name=None, overwrite=False, step=None):

        super().__init__()
        self.experiment = experiment

        self.figure_name = figure_name
        self.overwrite = overwrite
        self.step = step

    def forward(self, data):
        """
        Arguments
        ---------
        data: figure, optional.
            The figure you want to log. If `None` passed,
            the global pyplot figure will be logged and uploaded
        """
        self.experiment.log_figure(
            self.figure_name, data, self.overwrite, self.step,
        )
        return data


class Histogram3d(_base.Operation):
    """Logs a histogram of values for a 3D chart as an asset for this experiment.

    Calling this method multiple times with the same name and incremented steps
    will add additional histograms to the 3D chart on Comet.ml.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    name: str, optional
        Name of summary
    step: int, optional.
        Used as the Z axis when plotting on Comet.ml.
    **kwargs:
        Additional keyword arguments for histogram.

    Returns
    -------
    data: List | Tuple | Array | Histogram object
        Summarization of histogram (passed to `forward`).

    """

    def __init__(self, experiment, name=None, step=None, **kwargs):
        super().__init__()
        self.experiment = experiment

        self.name = name
        self.step = step
        self.kwargs = kwargs

    def forward(self, data):
        """
        Arguments
        ---------
        data: List | Tuple | Array | Histogram object
            Summarization of histogram
        """
        self.experiment.log_histogram_3d(data, self.name, self.step, **self.kwargs)
        return data


class Image(_base.Operation):
    """Logs the image. Images are displayed on the Graphics tab on Comet.ml.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    name: str, optional
        A custom name to be displayed on the dashboard.
        If not provided the filename from the image_data argument will be used if it is a path.
    overwrite: bool, optional
        If another image with the same name exists, it will be overwritten if overwrite is set to True.
    image_format: str, optional
        Default: 'png'. If the image_data is actually something that can be turned
        into an image, this is the format used. Typical values include 'png' and 'jpg'.
    image_scale: float, optional
        Default: 1.0. If the image_data is actually something that can be turned
        into an image, this will be the new scale of the image.
    image_shape: Tuple, optional
        Default: None. If the image_data is actually something that can be
        turned into an image, this is the new shape of the array. Dimensions are (width, height).
    image_colormap: str, optional
        If the image_data is actually something that can be turned into an image,
        this is the colormap used to colorize the matrix.
    image_minmax: (Number, Number), optional
        If the image_data is actually something that can be turned into an image,
        this is the (min, max) used to scale the values.
        Otherwise, the image is autoscaled between (array.min, array.max).
    image_channels: str, optional. Default 'last'.
        If the image_data is actually something that can be turned into an image,
        this is the setting that indicates where the color information is in the format of the 2D data.
        'last' indicates that the data is in (rows, columns, channels)
        where 'first' indicates (channels, rows, columns).
    copy_to_tmp: bool, optional
        If image_data is not a file path, then this flag determines if the image
        is first copied to a temporary file before upload.
        If copy_to_tmp is False, then it is sent directly to the cloud.
        Default: `True`
    step: int, optional
        Used to associate the audio asset to a specific step. Default: `None`

    Returns
    -------
    data: Multiple objects
        See `forward` for possibilities

    """

    def __init__(
        self,
        experiment,
        name=None,
        overwrite=False,
        image_format="png",
        image_scale=1.0,
        image_shape=None,
        image_colormap=None,
        image_minmax=None,
        image_channels="last",
        copy_to_tmp=True,
        step=None,
    ):
        super().__init__()

        self.experiment = experiment

        self.name = name
        self.overwrite = overwrite
        self.image_format = image_format
        self.image_scale = image_scale
        self.image_shape = image_shape
        self.image_colormap = image_colormap
        self.image_minmax = image_minmax
        self.image_channels = image_channels
        self.copy_to_tmp = copy_to_tmp
        self.step = step

    def forward(self, data):
        """
        Arguments
        ---------
        data: Multiple objects
            One of:
                * a path (string) to an image
                * a file-like object containing an image
                * a numpy matrix
                * a TensorFlow tensor
                * a PyTorch tensor
                * a list or tuple of values
                * a PIL Image

        """
        self.experiment.log_image(
            data,
            self.name,
            self.overwrite,
            self.image_format,
            self.image_scale,
            self.image_shape,
            self.image_colormap,
            self.image_minmax,
            self.image_channels,
            self.copy_to_tmp,
            self.step,
        )
        return data


class Scalar(_base.Operation):
    """Logs scalar value under specified name.

    Usually used to log metric values (like `accuracy`)

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    name: str
        Name of scalar / metric.
    step: int, optional
        Used as the X axis when plotting on comet.ml
    epoch: int, optional
        Used as the X axis when plotting on comet.ml
    include_context: bool, optional
        If set to True (the default), the current context will be logged along
        the metric.

    Returns
    -------
    data: Number
        Value passed to `forward`

    """

    def __init__(self, experiment, name, step=None, epoch=None, include_context=True):
        super().__init__()

        self.experiment = experiment

        self.name = name
        self.step = step
        self.epoch = epoch
        self.include_context = include_context

    def forward(self, data):
        """
        Arguments
        ---------
        data: Number
            Value to log
        """
        self.experiment.log_metric(
            self.name, data, self.step, self.epoch, self.include_context
        )
        return data


# log_metrics
class Scalars(_base.Operation):
    """Logs dictionary of scalars / metrics.

    Usually used to log metric values (like `accuracy`)

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    prefix: str, optional
        Name of prefix used when logging into comet.ml
    step: int, optional
        Used as the X axis when plotting on comet.ml
    epoch: int, optional
        Used as the X axis when plotting on comet.ml

    Returns
    -------
    data: Dict
        Dictionary passed to `forward`

    """

    def __init__(self, experiment, prefix=None, step=None, epoch=None):
        super().__init__()

        self.experiment = experiment

        self.prefix = prefix
        self.step = step
        self.epoch = epoch

    def forward(self, data):
        """
        Arguments
        ---------
        data: Dict
            Dictionary with values to log
        """
        self.experiment.log_metrics(data, self.prefix, self.step, self.epoch)
        return data


class Other(_base.Operation):
    """Reports a key and value to the Other tab on Comet.ml.

    Useful for reporting datasets attributes, datasets path, unique identifiers etc.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment

    Returns
    -------
    data: Tuple[Any, Any]
        Tuple with `key` and `value`

    """

    def __init__(self, experiment):
        super().__init__()
        self.experiment = experiment

    def forward(self, data):
        """
        Arguments
        ---------
        data: Tuple[Any, Any]
            Tuple with `key` and `value`
        """
        self.experiment.log_other(*data)
        return data


class Others(_base.Operation):
    """Reports dictionary of key/values to the Other tab on Comet.ml.

    Useful for reporting datasets attributes, datasets path, unique identifiers etc.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment

    Returns
    -------
    data: Dict[Any, Any]
        Dict with any `keys` and `values`

    """

    def __init__(self, experiment):
        super().__init__()
        self.experiment = experiment

    def forward(self, data):
        """
        Arguments
        ---------
        data: Dict[Any, Any]
            Dict with any `keys` and `values`
        """
        self.experiment.log_others(data)
        return data


class Table(_base.Operation):
    """Logs tabular data.

    These strings appear on the Text Tab in the Comet UI.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    filename: str
        Filename ending in ".csv", or ".tsv"
    headers: bool | List
        If True, will add column headers automatically if tabular_data is given;
        if False, no headers will be added; if list then it will be used as headers.

    Returns
    -------
    data: tensor | List[List]
        Data received in `forward`

    """

    def __init__(
        self,
        experiment,
        filename: str,
        headers: typing.Union[bool, typing.List] = False,
    ):
        super().__init__()
        self.experiment = experiment

        self.filename = filename
        self.header = headers

    def forward(self, data):
        """
        Arguments
        ---------
        data: tensor | List[List]
            Data that can be interpreted as 2D tabular data.

        """
        self.experiment.log_table(self.filename, data, self.header)
        return data


class Text(_base.Operation):
    """Logs the text.

    These strings appear on the Text Tab in the Comet UI.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    step: int, optional
        Used to associate text to a specific step
    metadata: JSON-like, optional
        Additional data attached to text.

    Returns
    -------
    data: str
        Received text

    """

    def __init__(self, experiment, step: int = None, metadata=None):
        super().__init__()
        self.experiment = experiment

        self.step = step
        self.metadata = metadata

    def forward(self, data):
        """
        Arguments
        ---------
        data: str
            Text to be stored

        """
        self.experiment.log_text(data, self.step, self.metadata)
        return data


class Notification(_base.Operation):
    """Send yourself a notification through email when an experiment ends.

    Parameters
    ----------
    experiment: Experiment
        Object representing single experiment
    title: str
        Subject of the email
    status: str
        Final status of the experiment.
        Typically, something like "finished", "completed" or "aborted".
    additional_data: dict
        Dictionary of key/values to notify.

    Returns
    -------
    data: Any
        Anything passed to forward

    """

    def __init__(self, experiment, title, status=None, additional_data=None):
        super().__init__()
        self.experiment = experiment

        self.title = title
        self.status = status
        self.additional_data = additional_data

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything as it's not send to the function, just passes through

        """
        self.experiment.log_text(self.title, self.status, self.additional_data)
        return data



import typing

import torch

from . import utils


def _to_tensor(value):
    if value is None or torch.is_tensor(value):
        return value
    return torch.tensor(value)


def _get_reduction(reduction):
    if reduction is None:
        return lambda loss: loss.sum() / loss.shape[0]
    return reduction


@utils.docs
def binary_focal_loss(
    outputs: torch.Tensor,
    targets: torch.Tensor,
    gamma: float,
    weight=None,
    pos_weight=None,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
) -> torch.Tensor:

    reduce = _get_reduction(reduction)
    weight, pos_weight = _to_tensor(weight), _to_tensor(pos_weight)

    probabilities = (1 - torch.sigmoid(outputs)) ** gamma
    loss = probabilities * torch.nn.functional.binary_cross_entropy_with_logits(
        outputs, targets.float(), weight, reduction="none", pos_weight=pos_weight,
    )

    return reduce(loss)


@utils.docs
def multiclass_focal_loss(
    outputs: torch.Tensor,
    targets: torch.Tensor,
    gamma: float,
    weight=None,
    ignore_index=-100,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
) -> torch.Tensor:

    reduce = _get_reduction(reduction)

    if ignore_index >= 0:
        outputs[:, ignore_index, ...] = 0
    probabilities = (1 - torch.nn.functional.softmax(outputs, dim=1)) ** gamma
    loss = probabilities * torch.nn.functional.cross_entropy(
        outputs, targets, weight, ignore_index=ignore_index, reduction="none"
    ).unsqueeze(dim=1)

    return reduce(loss)


@utils.docs
def smooth_binary_cross_entropy(
    outputs: torch.Tensor,
    targets: torch.Tensor,
    alpha: float,
    weight=None,
    pos_weight=None,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
) -> torch.Tensor:

    reduce = _get_reduction(reduction)

    weight, pos_weight = _to_tensor(weight), _to_tensor(pos_weight)
    targets = targets * (1 - alpha) + alpha / 2

    return reduce(
        torch.nn.functional.binary_cross_entropy_with_logits(
            outputs, targets, weight, pos_weight=pos_weight, reduction="none"
        )
    )


@utils.docs
def smooth_cross_entropy(
    outputs: torch.Tensor,
    targets: torch.Tensor,
    alpha: float,
    weight=None,
    ignore_index: int = -100,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
) -> torch.Tensor:

    reduce = _get_reduction(reduction)

    # All classes may not occur in loss, specify num_classes explicitly
    one_hot_targets = torch.nn.functional.one_hot(targets, num_classes=outputs.shape[1])
    permutation = list(range(len(one_hot_targets.shape)))
    permutation.insert(1, permutation[-1])

    # We leave last dimension as it's inserted at correct 1 position
    one_hot_targets = one_hot_targets.permute(permutation[:-1])

    smoothed_targets = one_hot_targets * (1 - alpha) + alpha / outputs.shape[1]
    if ignore_index >= 0:
        smoothed_targets[:, ignore_index, ...] = outputs[:, ignore_index, ...]
    log_softmax = torch.nn.functional.log_softmax(outputs, dim=1).unsqueeze(dim=1)
    loss = -(smoothed_targets * log_softmax)
    if weight is not None:
        loss *= weight.unsqueeze(dim=0)

    return reduce(loss)


@utils.docs
def quadruplet(
    anchor: torch.Tensor,
    positive: torch.Tensor,
    negative: torch.Tensor,
    negative2: torch.Tensor,
    alpha1: float = 1.0,
    alpha2: float = 0.5,
    metric: typing.Callable[
        [torch.Tensor, torch.Tensor], torch.Tensor
    ] = torch.nn.functional.pairwise_distance,
    weight=None,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = None,
) -> torch.Tensor:

    reduce = _get_reduction(reduction)

    loss = torch.nn.functional.relu(
        metric(anchor, positive) ** 2 - metric(anchor, negative) ** 2 + alpha1
    )
    +torch.nn.functional.relu(
        metric(anchor, positive) ** 2 - metric(negative, negative2) ** 2 + alpha2
    )

    if weight is not None:
        loss *= weight.unsqueeze(dim=0)

    return reduce(loss)



"""Module containing base classes shared amongst most (all?) classes in `torchtraining`.

Shouldn't be a concern to users, see docs of each if you want to help with
development.

"""

import abc
import functools
import inspect
import typing

import yaml

from . import exceptions


class Base:
    """Common base class for all `torchtraining` objects.

    Defines default `__str__` and `__repr__`.
    Most objects should customize `__str__` according to specific
    needs.

    Custom objects usually use `yaml.dump` to easily see parameters
    and whole pipeline.

    """

    def __str__(self) -> str:
        return f"{type(self).__module__}.{type(self).__name__}"

    def __repr__(self) -> str:
        parameters = ", ".join(
            "{}={}".format(key, value)
            for key, value in self.__dict__.items()
            if not key.startswith("_")
        )
        return "{}({})".format(self, parameters)


###############################################################################
#
#                                   PIPES
#
###############################################################################


class Pipe(Base):
    def __init__(self):
        self._pipes = []

    def _pipe(self, data, pipes):
        for pipe in pipes:
            data = pipe(data)
        return data

    def _broadcast(self, data, pipes):
        for pipe in pipes:
            pipe(data)
        return data

    def __pow__(self, other):
        if not isinstance(other, Pipe):
            raise ValueError(
                "Only `torchtraining.Pipe` objects can be used with ** chain operator."
                "Those objects are `torchtraining.Operation`, `torchtraining.iterations.IterationBase`, "
                "`torchtraining.Accumulator`, `torchtraining.steps.StepBase`, `torchtraining.epochs.EpochBase` "
                "or other inheriting from them. Did you mean to inherit from `torchtraining.Operation`?"
            )
        if isinstance(other, Producer):
            raise ValueError("Cannot pipe {} into {} as it's a Producer")
        if isinstance(other, Accumulator):
            if not isinstance(self, Producer):
                return other.add_previous(self)
        self._pipes.append(other)
        return self

    @abc.abstractmethod
    def __call__(self, data) -> typing.Any:
        pass

    @abc.abstractmethod
    def forward(self, data):
        pass


###############################################################################
#
#                                     OP
#
###############################################################################


class Operation(Pipe):
    """Base class for operations.

    Usually processes data returned / yielded by
    `torchtraining.steps.Step` / `torchtraining.iterations.Iteration` instances.

    Can also be used with `accumulators`, in this case those ops are used BEFORE
    data is passed to `accumulator`.

    Users should implement `forward` method which returns desired value
    after transformation.

    `__call__` does nothing, but is kept for API clarity.

    """

    def __call__(self, data) -> typing.Any:
        return self._pipe(self.forward(data), self._pipes)

    @abc.abstractmethod
    def forward(self, data) -> typing.Any:
        pass


###############################################################################
#
#                               ACCUMULATOR
#
###############################################################################


class Accumulator(Pipe):
    """Save values for further reuse.

    Values can be saved after each step or iteration.
    If any pipe is applied with `__gt__`, values will be first transformed
    by those pipes and saved as a last step.

    Users should implement `forward` method which returns value(s) which
    are saved after each step. As those values are only references,
    it is a low-cost operation.

    Users should also use `self.data` field in order to accumulate results
    somehow. See `torchtraining.accumulators` for examples.

    """

    def __init__(self):
        super().__init__()

        self.data = None
        self._previous = []

    def broadcast(self):
        """Run operations added AFTER accumulator.

        Used by `iterations` in order not to run `tensorboard` etc.
        at each value.

        """
        return self._broadcast(self.calculate(), self._pipes)

    def add_previous(self, pipe):
        self._previous.append(pipe)
        return self

    def __call__(self, data):
        return self.forward(self._pipe(data, reversed(self._previous)))

    @abc.abstractmethod
    def forward(self, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def calculate(self) -> typing.Any:
        pass

    @abc.abstractmethod
    def reset(self):
        pass


class Accelerator:
    """Base accelerator class.

    Currently IS NOT involved in any pipe'ing or operations
    (though this might be subject to change).

    Supports following as a syntactic sugar::

        accelerator ** step

    Always returns the passed object unchanged

    """

    def __pow__(self, other):
        return other


###############################################################################
#
#                                   PRODUCERS
#
###############################################################################


class Producer(Pipe):
    """Produce values and process/save them."""

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass


class ReturnProducer(Producer):
    """Producer returning values via `return` statement.

    Differentiation is needed for smooth user experience
    no matter generators (`iterations` or `epochs`) or simple returns
    (usually `step`).

    Applies pipes added via "**" and return UNMODIFIED data.

    """

    def __pow__(self, other):
        if isinstance(other, Accumulator):
            raise ValueError("You cannot accumulate values from steps.")
        return super().__pow__(other)

    def __call__(self, *args, **kwargs):
        data = self.forward(*args, **kwargs)
        self._broadcast(data, self._pipes)
        return data

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass


class GeneratorProducer(Producer):
    """Producer yielding values via `yield` statement.

    Differentiation is needed for smooth user experience
    no matter generators (`iterations` or `epochs`) or simple returns
    (usually `step`).

    Applies pipes added via "**" and yields UNMODIFIED data at each step.

    """

    def _reset_accumulators(self):
        for pipe in self._pipes:
            if isinstance(pipe, Accumulator):
                pipe.reset()

    def _broadcast_accumulators(self):
        for pipe in self._pipes:
            if isinstance(pipe, Accumulator):
                pipe.broadcast()

    def __call__(self, *args, **kwargs):
        for data in self.forward(*args, **kwargs):
            self._broadcast(data, self._pipes)
            yield data
        # Broadcast accumulators
        self._broadcast_accumulators()
        # Reset accumulators
        self._reset_accumulators()

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass


class Epoch(GeneratorProducer):
    """Base class for epoch-like structures.

    Works like GeneratorProducer, but handles exceptions.EpochsException
    in case earlier exit is required.

    Applies pipes added via "**" and yields UNMODIFIED data at each step.

    """

    def __call__(self, *args, **kwargs):
        with self:
            yield from super().__call__(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, _, exc_val, __):
        if isinstance(exc_val, exceptions.EpochsException):
            return True
        return False

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass

    def run(self, *args, **kwargs):
        for _ in self(*args, **kwargs):
            pass


class Iteration(GeneratorProducer):
    """Base class for iteration-like structures.

    Works like GeneratorProducer, but handles exceptions.IterationsException
    in case earlier exit is required from `iteration`.

    Applies pipes added via "**" and yields UNMODIFIED data at each step.

    """

    def __call__(self, *args, **kwargs):
        with self:
            yield from super().__call__(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, _, exc_val, __):
        if isinstance(exc_val, exceptions.IterationsException):
            return True
        return False

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass


class Step(ReturnProducer):
    """Base class for step-like structures.

    Works like ReturnProducer, is here to provide unified API across
    `steps`, `iterations` and `epochs`.
    """

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        pass



import torch
import torchtraining as tt
import torchvision

from . import data, iterations, modules, steps


def test_training():
    device = torch.device("cpu")

    dataloaders = data.get(batch_size=8)

    criterion = torch.nn.CrossEntropyLoss()
    network = modules.Classifier(in_channels=3, labels=100)
    optimizer = torch.optim.Adam(network.parameters())

    train_step, eval_step = (
        steps.train(optimizer, criterion, device),
        steps.eval(criterion, device),
    )

    writer = torch.utils.tensorboard.SummaryWriter("log")
    train_iteration, validation_iteration, test_iteration = (
        iterations.train(
            writer, train_step, network, dataloaders["train"], name="Train"
        ),
        iterations.eval(
            writer, eval_step, network, dataloaders["validation"], name="Validation"
        ),
        iterations.eval(writer, eval_step, network, dataloaders["test"], name="Test"),
    )

    epochs = tt.epochs.Epoch(
        train_iteration, validation_iteration, test_iteration, epochs=2
    )
    epochs.run()



"""Integrate `torchtraining` with `neptune.ai <https://neptune.ai/>`__ experiment management tool.

.. note::

    **IMPORTANT**: This module is experimental and may not be working
    correctly. Use at your own risk and report any issues you find.

.. note::

    **IMPORTANT**: This module needs `neptune-client` Python package to be available.
    You can install it with `pip install -U torchtraining[neptune]`

Usage is similar to `torchtraining.callbacks.Tensorboard`, except creating `project`
instead of `torch.utils.tensorboard.SummaryWriter`.

Example::

    import torchtraining as tt
    import torchtraining.callbacks.neptune as neptune


    class TrainStep(tt.steps.Train):
        def forward(self, module, sample):
            # Dummy step
            images, labels = sample
            return images


    project = neptune.project()
    experiment = neptune.experiment(project)

    step = TrainStep(criterion, device)

    # Experiment is optional
    # You have to split `tensor` as only single image can be logged
    # Also you need to cast to `numpy`
    step ** tt.OnSplittedTensor(tt.cast.Numpy() ** neptune.Image(experiment))


"""

import neptune

from .. import _base


def project(project_qualified_name=None, api_token=None, proxies=None, backend=None):
    """Initialize Neptune client library to work with specific project.

    Authorize user, sets value of global variable project to Project object
    that can be used to create or list experiments, notebooks, etc.

    Extensive documentation (explained optional parameters) is located
    `here <https://docs.neptune.ai/neptune-client/docs/neptune.html#neptune.init>`__ .

    Returns
    -------
    neptune.Project
        Object that is used to create or list experiments, notebooks, etc.

    """
    return neptune.init(project_qualified_name, api_token, proxies, backend)


def experiment(
    self,
    project=None,
    name=None,
    description=None,
    params=None,
    properties=None,
    tags=None,
    upload_source_files=None,
    abort_callback=None,
    logger=None,
    upload_stdout=True,
    upload_stderr=True,
    send_hardware_metrics=True,
    run_monitoring_thread=True,
    handle_uncaught_exceptions=True,
    git_info=None,
    hostname=None,
    notebook_id=None,
    notebook_path=None,
):
    """Create and start Neptune experiment.

    Create experiment, set its status to running and append it to the top of the experiments view.
    All parameters are optional.

    Extensive documentation (explained optional parameters) is located
    `here <https://docs.neptune.ai/neptune-client/docs/project.html#neptune.projects.Project.create_experiment>`__

    Returns:
    --------
        `neptune.experiments.Experiment` object that is used to manage experiment
        and log data to it. See `original documentation <https://docsneptune.ai/neptune-client/docs/experiment.html#neptune.experiments.Experiment.log_artifact>`__)

    Raises:
    ------
        `ExperimentValidationError`: When provided arguments are invalid.
        `ExperimentLimitReached`: When experiment limit in the project has been reached.

    """
    if project is None:
        return neptune.create_experiment(
            name,
            description,
            params,
            properties,
            tags,
            upload_source_files,
            abort_callback,
            logger,
            upload_stdout,
            upload_stderr,
            send_hardware_metrics,
            run_monitoring_thread,
            handle_uncaught_exceptions,
            git_info,
            hostname,
            notebook_id,
            notebook_path,
        )
    return project.create_experiment(
        name,
        description,
        params,
        properties,
        tags,
        upload_source_files,
        abort_callback,
        logger,
        upload_stdout,
        upload_stderr,
        send_hardware_metrics,
        run_monitoring_thread,
        handle_uncaught_exceptions,
        git_info,
        hostname,
        notebook_id,
        notebook_path,
    )


class _NeptuneOperation(_base.Operation):
    """Base class for neptune.ai operations.

    Based on `experiment` (and whether it's `None`) calls `neptune.method`
    or `experiment.method` (to be used in `forward`).

    """

    def __init__(self, experiment, method_name):
        super().__init__()
        self.experiment = experiment
        if self.experiment is None:
            self._method = getattr(neptune, method_name)
        else:
            self._method = getattr(self.experiment, method_name)


class Artifact(_NeptuneOperation):
    """Save an artifact (file) in experiment storage.

    Parameters
    ----------
    destination: str, optional
        Destination path to save artifact. If `None` artifact name will be used.
        Default: `None`
    experiment: `neptune.experiments.Experiment`, optional
        Instance of experiment to use. If `None`, global `experiment` will be used.
        Default: `None`


    Returns
    -------
    data
        Data without any modification

    """

    def __init__(self, destination=None, experiment=None):
        super().__init__(experiment, "send_artifact")
        self.destination = destination

    def forward(self, data):
        """
        Arguments
        ---------
        data: str | IO.Object
            A path to the file in local filesystem or IO object.
            It can be open file descriptor or in-memory buffer like io.StringIO or io.BytesIO.
        """
        self._method(data, self.destination)
        return data


class Image(_NeptuneOperation):
    """Log image data in Neptune.

    See `original documentation <https://docs.neptune.ai/_modules/neptune/experiments.html#Experiment.log_image>`__.

    Example::

        import torchtraining as tt
        import torchtraining.callbacks.neptune as neptune


        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                # Dummy step
                images, labels = sample
                # Images is of shape [batch, 1, 28, 28], say MNIST
                return images


        project = neptune.project()
        experiment = neptune.experiment(project)

        step = TrainStep(criterion, device)

        # Experiment is optional
        # You have to split `tensor` as only single image can be logged
        # Also you need to cast to `numpy`
        step ** tt.OnSplittedTensor(tt.cast.Numpy() ** neptune.Image(experiment))

    Parameters
    ----------
    log_name: str
        Name of log (group of images), e.g. "generated images".
    image_name: str, optional
        Name of this specific image received in `data`.
        If `None` consecutive numbers will be used.
        Default: `None`
    description: str, optional
        Textual description of image. If `None` no description.
        Default: `None`
    timestamp: time, optional
        Timestamp to be associated with log entry. Must be Unix time.
        If None is passed, time.time() (Python 3.6 example) is invoked to obtain timestamp.
        Default `None`
    experiment: `neptune.experiments.Experiment`, optional
        Instance of experiment to use. If `None`, global `experiment` will be used.
        Default: `None`

        Returns
    -------
    data
        Data without any modification

    """

    def __init__(
        self,
        log_name: str,
        image_name: str = None,
        description: str = None,
        timestamp=None,
        experiment=None,
    ):
        super().__init__(experiment, "send_image")
        self.log_name = log_name
        self.image_name = image_name
        self.description = description
        self.timestamp = timestamp

    def forward(self, data):
        """
        Arguments
        ---------
        data: PIL image | matplotlib.figure.Figure | str | np.array
            Can be one of:
            * :obj:`PIL image`
              `Pillow docs <https://pillow.readthedocs.io/en/latest/reference/Image.html#image-module>`_
            * :obj:`matplotlib.figure.Figure`
              `Matplotlib 3.1.1 docs <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.figure.Figure.html>`_
            * :obj:`str` - path to image file
            * 2-dimensional :obj:`numpy.array` - interpreted as grayscale image
            * 3-dimensional :obj:`numpy.array` - behavior depends on last dimension

                * if last dimension is 1 - interpreted as grayscale image
                * if last dimension is 3 - interpreted as RGB image
                * if last dimension is 4 - interpreted as RGBA image

            You may need to `transpose` and  transform PyTorch `tensors` to
            fit the above format.

        """
        self._method(
            self.log_name, data, None, self.image_name, self.description, self.timestamp
        )
        return data


class Scalar(_NeptuneOperation):
    """Log scalar data in Neptune.

    Calls `experiment.log_metric` under the hood.
    See `original documentation <https://docs.neptune.ai/_modules/neptune/experiments.html#Experiment.log_metric>`__.

    Example::

        import torchtraining as tt
        import torchtraining.callbacks.neptune as neptune


        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                # Dummy step
                # Calculate loss
                ...
                return loss


        project = neptune.project()
        experiment = neptune.experiment(project)

        step = TrainStep(criterion, device)

        # Experiment is optional
        # You have to split `tensor` as only single image can be logged
        # Also you need to cast to `numpy`
        step ** tt.cast.Item() ** neptune.Image(experiment))

    Parameters
    ----------
    log_name: str
        Name of log (group of images), e.g. "generated images".
    timestamp: time, optional
        Timestamp to be associated with log entry. Must be Unix time.
        If None is passed, time.time() (Python 3.6 example) is invoked to obtain timestamp.
        Default `None`
    experiment: `neptune.experiments.Experiment`, optional
        Instance of experiment to use. If `None`, global `experiment` will be used.
        Default: `None`

    Returns
    -------
    data
        Data without any modification

    """

    def __init__(
        self, log_name: str, timestamp=None, experiment=None,
    ):
        super().__init__(experiment, "send_metric")
        self.log_name = log_name
        self.timestamp = timestamp

    def forward(self, data):
        """
        Arguments
        ---------
        data: double
            Single element Python `double` type

        """
        self._method(self.log_name, data, None, self.timestamp)
        return data


class Text(_NeptuneOperation):
    """Log text data in Neptune.

    See `original documentation <https://docs.neptune.ai/_modules/neptune/experiments.html#Experiment.log_text>`__.

    Example::

        import torchtraining as tt
        import torchtraining.callbacks.neptune as neptune


        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                # You returning text for some reason...
                return loss, text


        project = neptune.project()
        neptune.experiment(project)

        step = TrainStep(criterion, device)

        # Experiment is optional
        # You have to split `tensor` as only single image can be logged
        # Also you need to cast to `numpy`
        step ** tt.Select(text=1) ** neptune.Text())

    Parameters
    ----------
    log_name: str
        Name of log (group of images), e.g. "generated images".
    timestamp: time, optional
        Timestamp to be associated with log entry. Must be Unix time.
        If None is passed, time.time() (Python 3.6 example) is invoked to obtain timestamp.
        Default `None`
    experiment: `neptune.experiments.Experiment`, optional
        Instance of experiment to use. If `None`, global `experiment` will be used.
        Default: `None`

    Returns
    -------
    data
        Data without any modification

    """

    def __init__(
        self, log_name: str, timestamp=None, experiment=None,
    ):
        super().__init__(experiment, "send_text")
        self.log_name = log_name
        self.timestamp = timestamp

    def forward(self, data):
        """
        Arguments
        ---------
        data: str
            Text to be logged
        """
        self._method(self.log_name, data, None, self.timestamp)
        return data


class Reset(_NeptuneOperation):
    """Resets the log.

    Removes all data from the log and enables it to be reused from scratch.
    See `original documentation <https://docs.neptune.ai/_modules/neptune/experiments.html#Experiment.reset_log>`__.

    Example::

        import torchtraining as tt
        import torchtraining.callbacks.neptune as neptune


        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                # You returning text for some reason...
                return loss, text


        project = neptune.project()
        neptune.experiment(project)

        step = TrainStep(criterion, device)

        # Experiment is optional
        # You have to split `tensor` as only single image can be logged
        # Also you need to cast to `numpy`
        step ** tt.Select(text=1) ** neptune.Text())

    Parameters
    ----------
    log_name: str
        Name of log (group of images), e.g. "generated images".
        If `log` does not exist, error `ChannelDoesNotExist` will be raised.

    Returns
    -------
    data
        Anything which was passed into it.

    """

    def __init__(
        self, log_name: str, experiment=None,
    ):
        super().__init__(experiment, "reset_log")
        self.log_name = log_name

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything as data will be forwarded
        """
        self._method(self.log_name)
        return data



"""Traditionally known callback-like pipes.

.. note::

    **IMPORTANT**: This module is one of core features
    so be sure to understand how it works.

This module allows user to (for example)::

    * `save` their best model
    * terminate training (early stopping)
    * log data to `stdout`

Example::

    class TrainStep(tt.steps.Train):
        def forward(self, module, sample):
            ...
            return loss, targets


    step = TrainStep(criterion, device)
    step ** tt.Select(loss=0) ** tt.callbacks.TerminateOnNan()

Users can also use specific `callbacks` which integrate with third party tools,
namely:

    * tensorboard
    * neptune
    * comet

.. note::

    **IMPORTANT**: Most of the training related logging/saving/processing
    is (or will be) in this package.

"""

import importlib
import numbers
import operator
import pathlib
import sys
import time
import typing

import loguru
import torch

from .. import _base, exceptions
from ..utils import general as utils

if utils.modules_exist("torch.utils.tensorboard"):
    from . import tensorboard


if utils.modules_exist("neptune"):
    from . import neptune


if utils.modules_exist("comet_ml"):
    from . import comet


class Save(_base.Operation):
    """Save best module according to specified metric.

    .. note::

        **IMPORTANT**: This class plays the role of `ModelCheckpointer`
        known from other training libs. It is user's role to load module
        and pass to `step`, hence we provide only `saving` part of checkpointing
        (may be subject to change).

    Example::

        import operator


        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss


        step = TrainStep(criterion, device)
        iteration = tt.iterations.Train(step, module, dataloader)

        # Lower (operator.lt) loss than current best -> save the model
        iteration ** tt.accumulators.Mean() ** tt.callbacks.Save(
            module, "my_model.pt", comparator=operator.lt
        )

    Parameters
    ----------
    module: torch.nn.Module
        Module to save.
    path: pathlib.Path
        Path where module will be saved. Usually ends with `.pt` suffix,
        see PyTorch documentation.
    comparator: Callable(Number, Number) -> bool, optional
        Function comparing two values - current metric and best metric.
        If ``true``, save new module and use current value as the best one.
        One can use Python's standard operator library for this argument.
        Default: `operator.gt` (`current` ** `best`)
    method: Callable(torch.nn.Module, pathlib.Path) -> None, optional
        Method to save `torch.nn.Module`. Takes module and path
        and returns anything (return value is discarded).
        Might be useful to transform model into
        `torch.jit.ScriptModule` or do some preprocessing before saving.
        Default: `torch.save` (whole model saving)
    log : str | int, optional
            Severity level for logging object's actions.
            Available levels of logging:
                * NONE      0
                * TRACE 	5
                * DEBUG 	10
                * INFO 	20
                * SUCCESS 	25
                * WARNING 	30
                * ERROR 	40
                * CRITICAL 	50
            Default: `INFO`


    Returns
    -------
    Any
        Data passed in forward

    """

    def __init__(
        self,
        module: torch.nn.Module,
        path: pathlib.Path,
        comparator: typing.Callable = operator.gt,
        method: typing.Callable = None,
        log: typing.Union[str, int] = "NONE",
    ):
        super().__init__()

        self.module = module
        self.path = path
        self.comparator = comparator
        self.method = (
            lambda module, path: torch.save(module, path) if method is None else method
        )
        self.log = log

        self.best = None

    def forward(self, data: typing.Any) -> typing.Any:
        """
        Arguments
        ---------
        data: Any
            Anything which can be passed to `comparator` (e.g. `torch.Tensor`).
        """
        if self.best is None or self.comparator(data, self.best):
            self.best = data
            self.method(self.module, self.path)
            loguru.logger.log(
                self.log, "New best value: {}".format(self.best),
            )

        return data


class TimeStopping(_base.Operation):
    """Stop `epoch` after specified duration.

    Python's `time.time()` functionality is used.

    Can be placed anywhere (e.g. `step ** TimeStopping(60 * 60)`) as it's
    not data dependent.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss


        step = TrainStep(criterion, device)
        iteration = tt.iterations.Train(step, module, dataloader)

        # Stop after 30 minutes
        iteration ** tt.callbacks.TimeStopping(duration=60 * 30)

    Parameters
    ----------
    duration: int | float
        How long to run (in seconds) before exiting program.
    log : str | int, optional
            Severity level for logging object's actions.
            Available levels of logging:
                * NONE      0
                * TRACE 	5
                * DEBUG 	10
                * INFO 	20
                * SUCCESS 	25
                * WARNING 	30
                * ERROR 	40
                * CRITICAL 	50
            Default: `INFO`


    Returns
    -------
    Any
        Data passed in forward

    """

    def __init__(
        self, duration: float, log="NONE",
    ):
        super().__init__()

        self.duration = duration
        self.log = log
        self._start = time.time()

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything as `data` will be simply forwarded
        """
        if time.time() - self._start ** self.duration:
            loguru.logger.log(
                self.log, "Stopping after {} seconds.".format(self.duration)
            )
            raise exceptions.TimeStopping()
        return data


class TerminateOnNan(_base.Operation):
    """Stop `epoch` if any `NaN` value encountered in `data`.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss, targets


        step = TrainStep(criterion, device)
        step ** tt.Select(loss=0) ** tt.callbacks.TerminateOnNan()

    Parameters
    ----------
    log : str | int, optional
            Severity level for logging object's actions.
            Available levels of logging:
                * NONE      0
                * TRACE 	5
                * DEBUG 	10
                * INFO 	20
                * SUCCESS 	25
                * WARNING 	30
                * ERROR 	40
                * CRITICAL 	50
            Default: `INFO`

    Returns
    -------
    Any
        Data passed in forward

    """

    def __init__(
        self, log: typing.Union[str, int] = "NONE",
    ):
        super().__init__()

        self.log = log

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor possibly containing `NaN` values.
        """
        if torch.any(torch.isnan(data)):
            loguru.logger.log(self.log, "NaN values found, exiting with 1.")
            raise exceptions.TerminateOnNan()
        return data


class EarlyStopping(_base.Operation):
    """Stop `epoch` if `patience` was reached without improvement.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss, accuracy


        step = TrainStep(criterion, device)
        iteration = tt.iterations.Train(step, module, dataloader)

        # Stop if mean accuracy did not improve for `5` iterations
        iteration ** tt.Select(accuracy=1) ** tt.accumulators.Mean() ** tt.callbacks.EarlyStopping(
            patience=5
        )
        # Assume epoch was created from `iteration`

    Parameters
    ----------
    patience: int
        How long not to terminate if metric does not improve
    delta: Number, optional
        Difference between `best` value and current considered as an improvement.
        Default: `0`.
    comparator: Callable(Number, Number) -> bool, optional
        Function comparing two values - current metric and best metric.
        If ``true``, reset patience and use current value as the best one.
        One can use Python's standard `operator` library for this argument.
        Default: `operator.gt` (`current` ** `best`)
    log : str | int, optional
        Severity level for logging object's actions.
        Available levels of logging:
            * NONE      0
            * TRACE 	5
            * DEBUG 	10
            * INFO 	20
            * SUCCESS 	25
            * WARNING 	30
            * ERROR 	40
            * CRITICAL 	50
        Default: `INFO`

    Returns
    -------
    Any
        Data passed in forward

    """

    def __init__(
        self,
        patience: int,
        delta: numbers.Number = 0,
        comparator: typing.Callable = operator.gt,
        log="NONE",
    ):
        super().__init__()

        self.patience = patience
        self.delta = delta
        self.comparator = comparator
        self.log = log
        self.best = None

        self._counter = -1

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything which can be passed to `comparator` (e.g. `torch.Tensor`).
        """
        if self.best is None or self.comparator(data, self.best):
            self._counter = -1
        else:
            self._counter += 1
        if self._counter == self.patience:
            loguru.logger.log(
                self.log, "Stopping early, best found: {}".format(self.best)
            )
            raise exceptions.EarlyStopping()


class Unfreeze(_base.Operation):
    """Unfreeze module's parameters after `n` steps.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss, accuracy


        step = TrainStep(criterion, device)
        iteration = tt.iterations.Train(step, module, dataloader)

        # Assume `module`'s parameters are frozen

        # Doesn't matter what data goes it, so you can unfreeze however you wish
        # And it doesn't matter what the accumulated value is
        iteration ** tt.Select(accuracy=1) ** tt.accumulators.Sum() ** tt.callbacks.Unfreeze(
            module
        )

    Parameters
    ----------
    module: torch.nn.Module
        Module whose `parameters` will be unfrozen (`grad` set to `True`).
    n: int
        Module will be unfrozen after this many steps.
    log : str | int, optional
        Severity level for logging object's actions.
        Available levels of logging:
            * NONE      0
            * TRACE 	5
            * DEBUG 	10
            * INFO 	20
            * SUCCESS 	25
            * WARNING 	30
            * ERROR 	40
            * CRITICAL 	50
        Default: `INFO`


    Returns
    -------
    Any
        Data passed in forward

    """

    def __init__(self, module, n: int = 0, log="NONE"):
        super().__init__()

        self.module = module
        self.n = n
        self.log = log

        self._counter = -1

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything as data is simply forwarded

        """
        self._counter += 1
        if self._counter == self.n:
            loguru.logger.log(self.log, "Unfreezing module's parameters")
            for param in self.module.parameters():
                param.requires_grad_(True)
        return data


class Log(_base.Operation):
    r"""Log data using `loguru.logger`.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss, accuracy


        step = TrainStep(criterion, device)
        iteration = tt.iterations.Train(step, module, dataloader)

        # Log with loguru.logger accuracy
        iteration ** tt.Select(accuracy=1) ** tt.callbacks.Logger("Accuracy")

    Parameters
    ----------
    name : str
        Name under which data will be logged.
        It will be in format "{name}: {data}"
    log : str | int, optional
        Severity level for logging object's actions.
        Available levels of logging:
            * NONE      0
            * TRACE 	5
            * DEBUG 	10
            * INFO 	20
            * SUCCESS 	25
            * WARNING 	30
            * ERROR 	40
            * CRITICAL 	50
        Default: `INFO`

    Arguments
    ---------
    data: Any
        Anything which can be sensibly represented with `__str__` magic method.

    Returns
    -------
    Any
        Data passed in forward

    """

    def __init__(self, name: str, log="INFO"):
        super().__init__()

        self.name = name
        self.log = log

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything which can be sensibly represented with `__str__` magic method.
        """

        loguru.logger.log(self.log, "{}: {}".format(self.name, data))
        return data



"""Accumulate results from `iterations` or `epochs`

.. note::

    **IMPORTANT**: This module is one of core features
    so be sure to understand how it works.

.. note::

    **IMPORTANT**: Accumulators should be applied to `iteration`
    objects. This way those can efficiently accumulate value later passed
    to other operations.

Example::

    iteration
    ** tt.Select(predictions=1, labels=2)
    ** tt.metrics.classification.multiclass.Accuracy()
    ** tt.accumulators.Mean()
    ** tt.Split(
        tt.callbacks.Log(f"{name} Accuracy"),
        tt.callbacks.tensorboard.Scalar(writer, f"{name}/Accuracy"),
    )

Code above will accumulate `accuracy` from each step and after `iteration`
ends it will be send to `tt.Split`.


.. note::

    **IMPORTANT**: If users wish to implement their own `accumulators`
    `forward` shouldn't return anything but accumulate data in `self.data`
    variable. No argument `calculate` should return `self.data` after
    calculating accumulated value (e.g. for `mean` it would be division
    by number of samples).

"""

import abc
import typing

import torch

from ._base import Accumulator


class Sum(Accumulator):
    """Sum data coming into this object.

    `data` should have `+=` operator implemented between it's instances
    and Python integers.

    .. note::

        **IMPORTANT**: This is one of memory efficient accumulators
        and can be safely used.

    Returns
    -------
    torch.Tensor | Any
        Sum of values after accumulation. At each step proper
        summation up to this point is returned nonetheless.
        `torch.Tensor` usually, but can be anything "summable".

    """

    def __init__(self):
        super().__init__()
        self.data = 0

    def reset(self) -> None:
        """Assign 0 to `self.data` clearing `saver`."""
        self.data = 0

    def forward(self, data) -> None:
        """
        Arguments
        ---------
        data: Any
            Anything which has `__iadd__`/`__add__` operator implemented between it's instances
            and Python integers.
        """
        self.data += data

    def calculate(self) -> typing.Any:
        """Calculate final value.

        Returns
        -------
        torch.Tensor
            Data accumulated via addition.

        """
        return self.data


class Mean(Accumulator):
    """Take mean of the data coming into this object.

    `data` should have `+=` operator implemented between it's instances
    and Python integers.

    .. note::

        **IMPORTANT**: This is one of memory efficient accumulators
        and can be safely used. Should be preferred over accumulating
        data via `torchtraining.accumulators.List`


    Returns
    -------
    torch.Tensor | Any
        Mean of values after accumulation. At each step proper
        mean up to this point is returned nonetheless.
        `torch.Tensor` usually, but can be anything implementing concept above.

    """

    def __init__(self):
        super().__init__()
        self.data = 0
        self._counter = 0

    def reset(self) -> None:
        """Assign `0` to `self.data` and zero out counter clearing `saver`"""
        self.data = 0
        self._counter = 0

    def forward(self, data: typing.Any) -> None:
        """
        Arguments
        ---------
        data: Any
            Anything which has `__iadd__`/`__add__` operator implemented between it's instances
            and Python integers. It should also have `__div__` operator implemented
            for proper mean calculation.
        """
        self._counter += 1
        self.data += data

    def calculate(self) -> typing.Any:
        """Calculate final value.

        Returns
        -------
        torch.Tensor
            Accumulated data after summation and division by number of samples.

        """
        return self.data / self._counter


class List(Accumulator):
    """Sum data coming into this object.

    .. note::

        **IMPORTANT**: It is advised **NOT TO USE** this accumulator
        due to memory inefficiencies. Prefer `torchtraining.accumulators.Sum`
        or `torchtraining.accumulators.Mean` instead.

    List containing data received up to this moment.
    `data` **does not** have to implement any concept
    (as it is only appended to `list`).


    Returns
    -------
    List
        List of values after accumulation. At each step proper
        `list` up to this point is returned nonetheless.

    """

    def __init__(self):
        super().__init__()
        self.data = []

    def reset(self) -> None:
        """Assign empty `list` to `self.data clearing `saver`"""
        self.data = []

    def forward(self) -> typing.List[typing.Any]:
        """
        Arguments
        ---------
        data: Any
            Anything which can be added to `list`. So anything I guess
        """
        return self.data

    def accumulate(self, data) -> None:
        """Calculate final value.

        Returns
        -------
        torch.Tensor
            Return `List` with gathered data.

        """
        self.data.append()


class Except(Accumulator):
    """Special modifier of accumulators accumulating every value except specified.

    .. note::

        **IMPORTANT**: One of the `begin`, `end` has to be specified.

    .. note::

        **IMPORTANT**: This accumulators is useful in conjunction
        with `torchtraining.iterations.Multi` (e.g. for GANs and other irregular
        type of training).

    User can effectively choose which data coming from step should be accumulated
    and can divide accumulation based on that.

    Parameters
    ----------
    accumulator: tt.Accumulator
        Instance of accumulator to use for `data` accumulation.
    begin: int | torch.Tensor[int], optional
        If `int`, it should specify beginning of incoming values stream which
        will not be taken into accumulation. If `torch.Tensor` containing integers,
        it should specify consecutive beginnings of streams which are not taken into account.
        If left unspecified (`None`), `begin` is assumed to be `0`th step.
        Every modulo element of stream matching [begin, end] range will not be
        forwarded to accumulator.
    end: int | torch.Tensor[int], optional
        If `int`, it should specify end of incoming values stream which
        will not be taken into accumulation. If `torch.Tensor` containing integers,
        it should specify consecutive ends of stream which will not be taken into account.
        If left unspecified (`None`), `end` is assumed to be the same as `begin`.
        This effectively excludes every `begin` element coming from value stream.
        Every modulo element of stream matching [begin, end] range will not be
        forwarded to accumulator.


    Returns
    -------
    Any
        Whatever `accumulator` returns after accumulation. At each step proper
        value up to this point is returned nonetheless. Usually `torch.Tensor`
        or `list`.

    """

    def __init__(
        self, accumulator: Accumulator, begin=None, end=None,
    ):
        def _validate_argument(arg, name: str):
            if torch.is_tensor(arg):
                if not arg.dtype in (
                    torch.uint8,
                    torch.int8,
                    torch.int16,
                    torch.int32,
                    torch.int64,
                ):
                    raise ValueError(
                        "{} as torch.Tensor should be of integer-like type, got: {}".format(
                            name, arg.dtype
                        )
                    )
            if not isinstance(arg, int):
                raise ValueError(
                    "{} should be torch.Tensor/int, got : {}".format(name, type(arg))
                )
            return arg

        def _same_type(arg1, arg2):
            if not isinstance(arg1, type(arg2)):
                raise ValueError(
                    "begin and end should be of the same type. Got {} for begin and {} for end".format(
                        type(arg1), type(arg2)
                    )
                )

        def _same_shape(arg1, arg2):
            if arg1.shape != arg2.shape:
                raise ValueError(
                    "begin and end should have the same shape. Got {} for begin and {} for end".format(
                        arg1.shape, arg2.shape
                    )
                )

        def _one_of_specified(arg1, arg2):
            if arg1 is None and arg2 is None:
                raise ValueError("One of begin/end has to be specified.")

        super().__init__()

        _one_of_specified(begin, end)
        self.begin = _validate_argument(begin, "begin")
        if end is None:
            self.end = self.begin
        else:
            _same_type(self.begin, end)
            self.end = _validate_argument(end, "end")
            _same_shape(begin, end)

        self.accumulator = accumulator
        if isinstance(self.begin, int):
            self._in = lambda i: (self.begin <= (i % self.end + 1) <= self.end) and (
                i != self.end + 1
            )
        else:
            maximum = torch.max(self.end)
            self._in = lambda i: (
                (self.begin <= (i % maximum + 1)) & ((i % maximum + 1) <= self.end)
            ).any() & (i != maximum + 1)

        self._counter = -1

    def reset(self) -> None:
        """Reset internal `accumulator`."""
        self._counter = -1
        self.accumulator.reset()

    def forward(self, data) -> None:
        """
        Arguments
        ---------
        data: Any
            Anything which `accumulator` can consume
        """
        self._counter += 1
        if not self._in(self._counter):
            self.accumulator(data)

    def calculate(self) -> typing.Any:
        """Calculate final value.

        Returns
        -------
        Any
            Returns anything `accumulator` accumulated.

        """
        return self.accumulator.calculate()



import typing

import torch

from ... import utils
from . import utils as multiclass_utils


@utils.docs
def topk(
    output: torch.Tensor,
    target: torch.Tensor,
    k: int,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor] = torch.mean,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)

    biggest_indices = torch.topk(output, k, dim=-1)[1]
    equal = target.expand(*(target.shape), k) == biggest_indices

    return reduction(equal.sum(dim=-1))


@utils.docs
def accuracy(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor] = torch.mean,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.categorical(output, target)

    return reduction((output == target).float())


@utils.docs
def jaccard(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.mean,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    union = (output | target).sum(axis=-1)
    intersection = (target & output).sum(axis=-1)
    empty = union <= 0
    union[empty] = 1
    intersection[empty] = 1

    return reduction(intersection.float() / union)


# Basic cases


@utils.docs
def true_positive(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return reduction((output & target).float(), dim=-1)


@utils.docs
def false_positive(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return reduction((output & ~target).float(), dim=-1)


@utils.docs
def true_negative(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return reduction((~output & ~target).float(), dim=-1)


@utils.docs
def false_negative(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return reduction((~output & target).float(), dim=-1)


@utils.docs
def confusion_matrix(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    num_classes = output.shape[1]
    output, target = multiclass_utils.multiclass.categorical(output, target)

    unique_labels = (target * num_classes + output).flatten()

    return torch.bincount(unique_labels, minlength=num_classes ** 2).reshape(
        num_classes, num_classes
    )


@utils.docs
def recall(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return (output & target).sum().float() / target.sum()


@utils.docs
def specificity(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )
    inverse_target = ~target

    return (~output & inverse_target).sum().float() / inverse_target.sum()


@utils.docs
def precision(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return (output & target).sum().float() / output.sum()


@utils.docs
def negative_predictive_value(
    output: torch.Tensor, target: torch.Tensor,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )
    inverse_output = ~output

    return (inverse_output & ~target).sum().float() / inverse_output.sum()


@utils.docs
def false_negative_rate(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return (~output & target).sum().float() / target.sum()


@utils.docs
def false_positive_rate(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )
    inverse_target = ~target

    return (output & inverse_target).sum().float() / inverse_target.sum()


@utils.docs
def false_discovery_rate(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    return (output & ~target).sum().float() / output.sum()


@utils.docs
def false_omission_rate(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )
    inverse_output = ~output

    return (inverse_output & target).sum().float() / inverse_output.sum()


@utils.docs
def critical_success_index(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    tp = (output & target).sum().float()

    return tp / tp + (output != target).sum()


@utils.docs
def balanced_accuracy(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )
    inverse_target = ~target

    return (
        (output & target).sum().float() / target.sum()
        + (~output & inverse_target).sum().float() / inverse_target.sum()
    ) / 2


@utils.docs
def f1(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    tp = 2 * (output & target).sum().float()

    return tp / (tp + (output != target).sum())


@utils.docs
def f_beta(output: torch.Tensor, target: torch.Tensor, beta: float) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )

    tp = (1 + beta) ** 2 * (output & target).sum().float()

    return tp / (tp + (beta ** 2) * (output != target).sum())


@utils.docs
def matthews_correlation_coefficient(
    output: torch.Tensor, target: torch.Tensor,
) -> torch.Tensor:
    multiclass_utils.multiclass.check(output, target)
    output, target = multiclass_utils.multiclass.one_hot(
        output, target, num_classes=output.shape[1]
    )
    inverse_output = ~output
    inverse_target = ~target

    dimensions = tuple(range(len(inverse_output.shape)))

    tp = (output & target).float().sum(dim=dimensions[1:])
    tn = (inverse_output & inverse_target).float().sum(dim=dimensions[1:])
    fp = (output & inverse_target).float().sum(dim=dimensions[1:])
    fn = (inverse_output & target).float().sum(dim=dimensions[1:])

    numerator = torch.dot(tp, tn) - torch.dot(fp, fn)
    denominator = (
        output.sum() * target.sum() * inverse_target.sum() * inverse_output.sum()
    )

    if denominator == 0.0:
        return numerator
    return numerator / denominator
